# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for license terms.

from collections.abc import Generator
from typing import Any

import simpy as SIM

import upstage_des.api as UP
from upstage_des.task import InterruptStates
from upstage_des.type_help import SIMPY_GEN, TASK_GEN


class Cashier(UP.Actor):
    scan_speed = UP.State[float](
        valid_types=(float,),
        frozen=True,
    )
    time_until_break = UP.State[float](
        default=120.0,
        valid_types=(float,),
        frozen=True,
    )
    breaks_until_done = UP.State[int](default=2, valid_types=int)
    breaks_taken = UP.State[int](default=0, valid_types=int, recording=True)
    items_scanned = UP.State[int](
        default=0,
        valid_types=(int,),
        recording=True,
    )
    time_scanning = UP.LinearChangingState(
        default=0.0,
        valid_types=(float,),
    )
    messages = UP.ResourceState[UP.SelfMonitoringStore](
        default=UP.SelfMonitoringStore,
    )

    def time_left_to_break(self) -> float:
        elapsed = self.env.now - float(self.get_knowledge("start_time", must_exist=True))
        return self.time_until_break - elapsed


class CheckoutLane(UP.Actor):
    customer_queue = UP.ResourceState[UP.SelfMonitoringStore](
        default=UP.SelfMonitoringStore,
    )


class StoreBoss(UP.UpstageBase):
    def __init__(self, lanes: list[CheckoutLane]) -> None:
        self.lanes = lanes
        self._lane_map: dict[CheckoutLane, Cashier] = {}

    def get_lane(self, cashier: Cashier) -> CheckoutLane:
        possible = [lane for lane in self.lanes if lane not in self._lane_map]
        lane = self.stage.random.choice(possible)
        self._lane_map[lane] = cashier
        return lane

    def clear_lane(self, cashier: Cashier) -> None:
        to_del = [name for name, cash in self._lane_map.items() if cash is cashier]
        for name in to_del:
            del self._lane_map[name]


class CashierBreakTimer(UP.Task):
    def task(self, *, actor: Cashier) -> TASK_GEN:
        yield UP.Wait(actor.time_until_break)
        actor.interrupt_network("CashierJob", cause=dict(reason="BREAK TIME"))


class InterruptibleTask(UP.Task):
    def on_interrupt(self, *, actor: Cashier, cause: dict[str, Any]) -> InterruptStates:
        # We will only interrupt with a dictionary of data
        assert isinstance(cause, dict)
        job_list: list[str]

        if cause["reason"] == "BREAK TIME":
            job_list = ["Break"]
        elif cause["reason"] == "NEW JOB":
            job_list = cause["job_list"]
        else:
            raise UP.SimulationError("Unexpected interrupt cause")

        # determine time until break
        time_left = actor.time_left_to_break()
        # if there are only five minutes left, take the break and queue the task.
        if time_left <= 5.0 and "Break" not in job_list:
            job_list = ["Break"] + job_list

        # Ignore the interrupt, unless we've marked it to know otherwise
        marker = self.get_marker() or "none"
        if marker == "on break":
            if "Break" in job_list:
                job_list.remove("Break")

        self.clear_actor_task_queue(actor)
        self.set_actor_task_queue(actor, job_list)
        if marker == "cancellable":
            return self.INTERRUPT.END
        return self.INTERRUPT.IGNORE


class GoToWork(UP.Task):
    def task(self, *, actor: Cashier) -> TASK_GEN:
        """Go to work"""
        yield UP.Wait(15.0)


class TalkToBoss(UP.DecisionTask):
    def make_decision(self, *, actor: Cashier) -> None:
        """Zero-time task to get information."""
        boss: StoreBoss = self.stage.boss
        lane = boss.get_lane(actor)
        self.set_actor_knowledge(actor, "checkout_lane", lane, overwrite=False)
        actor.breaks_taken = 0
        self.set_actor_knowledge(actor, "start_time", self.env.now, overwrite=True)
        # Convenient spot to run the timer.
        CashierBreakTimer().run(actor=actor)


class WaitInLane(InterruptibleTask):
    def task(self, *, actor: Cashier) -> TASK_GEN:
        """Wait until break time, or a customer."""
        lane: CheckoutLane = self.get_actor_knowledge(
            actor,
            "checkout_lane",
            must_exist=True,
        )
        customer_arrival = UP.Get(lane.customer_queue)

        self.set_marker(marker="cancellable")
        yield customer_arrival

        customer: int = customer_arrival.get_value()
        self.set_actor_knowledge(actor, "customer", customer, overwrite=True)


class DoCheckout(InterruptibleTask):
    def task(self, *, actor: Cashier) -> TASK_GEN:
        """Do the checkout"""
        items: int = self.get_actor_knowledge(
            actor,
            "customer",
            must_exist=True,
        )
        per_item_time = actor.scan_speed / items
        actor.activate_linear_state(
            state="time_scanning",
            rate=1.0,
            task=self,
        )
        for _ in range(items):
            yield UP.Wait(per_item_time)
            actor.items_scanned += 1
        actor.deactivate_all_states(task=self)
        # assume 2 minutes to take payment
        yield UP.Wait(2.0)


class Break(UP.DecisionTask):
    def make_decision(self, *, actor: Cashier) -> None:
        """Decide what kind of break we are taking."""
        actor.breaks_taken += 1

        # we might have jobs queued
        queue = self.get_actor_task_queue(actor) or []
        if "Break" in queue:
            raise UP.SimulationError("Odd task network state")
        self.clear_actor_task_queue(actor)

        if actor.breaks_taken == actor.breaks_until_done:
            self.set_actor_task_queue(actor, ["NightBreak"])
        elif actor.breaks_taken > actor.breaks_until_done:
            raise UP.SimulationError("Too many breaks taken")
        else:
            self.set_actor_task_queue(actor, ["ShortBreak"] + queue)


class ShortBreak(InterruptibleTask):
    def task(self, *, actor: Cashier) -> TASK_GEN:
        """Take a short break."""
        self.set_marker("on break")
        yield UP.Wait(15.0)
        self.set_actor_knowledge(actor, "start_time", self.env.now, overwrite=True)
        CashierBreakTimer().run(actor=actor)


class NightBreak(UP.Task):
    def task(self, *, actor: Cashier) -> TASK_GEN:
        """Go home and rest."""
        self.clear_actor_knowledge(actor, "checkout_lane")
        self.stage.boss.clear_lane(actor)
        yield UP.Wait(60 * 12.0)


class Restock(UP.Task):
    def task(self, *, actor: Cashier) -> TASK_GEN:
        """Restock."""
        yield UP.Wait(10.0)


task_classes = {
    "GoToWork": GoToWork,
    "TalkToBoss": TalkToBoss,
    "WaitInLane": WaitInLane,
    "DoCheckout": DoCheckout,
    "Break": Break,
    "ShortBreak": ShortBreak,
    "NightBreak": NightBreak,
    "Restock": Restock,
}

task_links = {
    "GoToWork": UP.TaskLinks(default="TalkToBoss", allowed=["TalkToBoss"]),
    "TalkToBoss": UP.TaskLinks(default="WaitInLane", allowed=["WaitInLane"]),
    "WaitInLane": UP.TaskLinks(default="DoCheckout", allowed=["DoCheckout", "Break"]),
    "DoCheckout": UP.TaskLinks(default="WaitInLane", allowed=["WaitInLane", "Break"]),
    "Break": UP.TaskLinks(default="ShortBreak", allowed=["ShortBreak", "NightBreak"]),
    "ShortBreak": UP.TaskLinks(default="WaitInLane", allowed=["WaitInLane"]),
    "NightBreak": UP.TaskLinks(default="GoToWork", allowed=["GoToWork"]),
    "Restock": UP.TaskLinks(default="WaitInLane", allowed=["WaitInLane", "Break"]),
}

cashier_task_network = UP.TaskNetworkFactory(
    name="CashierJob",
    task_classes=task_classes,
    task_links=task_links,
)


class CashierMessages(UP.Task):
    def task(self, *, actor: Cashier) -> TASK_GEN:
        getter = UP.Get(actor.messages)
        yield getter
        tasks_needed: list[str] | str = getter.get_value()
        tasks_needed = [tasks_needed] if isinstance(tasks_needed, str) else tasks_needed
        actor.interrupt_network("CashierJob", cause=dict(reason="NEW JOB", job_list=tasks_needed))


cashier_message_net = UP.TaskNetworkFactory.from_single_looping("Messages", CashierMessages)


def customer_spawner(
    env: SIM.Environment,
    lanes: list[CheckoutLane],
) -> Generator[SIM.Event, None, None]:
    # sneaky way to get access to stage
    stage = lanes[0].stage
    while True:
        hrs = env.now / 60
        time_of_day = hrs // 24
        if time_of_day <= 8 or time_of_day >= 15.5:
            time_until_open = (24 - time_of_day) + 8
            yield env.timeout(time_until_open)

        lane_pick = stage.random.choice(lanes)
        number_pick = stage.random.randint(3, 17)
        yield lane_pick.customer_queue.put(number_pick)
        yield UP.Wait.from_random_uniform(5.0, 30.0).as_event()


def manager_process(boss: StoreBoss, cashiers: list[Cashier]) -> SIMPY_GEN:
    while True:
        # Use the random uniform feature, but convert the UPSTAGE event to simpy
        # because this is a simpy only process
        yield UP.Wait.from_random_uniform(30.0, 90.0).as_event()
        possible = [
            cash
            for cash in cashiers
            if getattr(cash.get_running_task("CashierJob"), "name", "") != "NightBreak"
        ]
        if not possible:
            return
        cash = boss.stage.random.choice(possible)
        yield cash.messages.put(["Restock"])


def test_cashier_example() -> None:
    with UP.EnvironmentContext(initial_time=8 * 60) as env:
        UP.add_stage_variable("time_unit", "min")
        cashier = Cashier(
            name="Bob",
            scan_speed=1.0,
            time_until_break=120.0,
            breaks_until_done=4,
            debug_log=True,
        )
        lane_1 = CheckoutLane(name="Lane 1")
        lane_2 = CheckoutLane(name="Lane 2")
        boss = StoreBoss(lanes=[lane_1, lane_2])

        UP.add_stage_variable("boss", boss)

        net = cashier_task_network.make_network()
        cashier.add_task_network(net)
        cashier.start_network_loop(net.name, "GoToWork")

        net = cashier_message_net.make_network()
        cashier.add_task_network(net)
        cashier.start_network_loop(net.name, "CashierMessages")

        customer_proc = customer_spawner(env, [lane_1, lane_2])
        _ = env.process(customer_proc)

        _ = env.process(manager_process(boss, [cashier]))

        env.run(until=20 * 60)

    for line in cashier.get_log():
        if "Interrupt" in line:
            print(line)

    print(cashier.items_scanned)


if __name__ == "__main__":
    test_cashier_example()



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""A state defines the conditions of an actor over time."""

from collections.abc import Callable
from copy import deepcopy
from enum import Enum
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

from simpy import Container, Store

from upstage_des.base import SimulationError, UpstageError
from upstage_des.data_types import CartesianLocation, GeodeticLocation
from upstage_des.math_utils import _vector_add, _vector_subtract
from upstage_des.resources.monitoring import SelfMonitoringStore
from upstage_des.task import Task

if TYPE_CHECKING:
    from upstage_des.actor import Actor

__all__ = (
    "ActiveState",
    "State",
    "LinearChangingState",
    "CartesianLocationChangingState",
    "GeodeticLocationChangingState",
    "ResourceState",
    "DetectabilityState",
    "CommunicationStore",
)

CALLBACK_FUNC = Callable[["Actor", Any], None]
ST = TypeVar("ST")


class ActiveStatus(Enum):
    activating = "ACTIVATING"
    deactivating = "DEACTIVATING"


def _compare(a: Any, b: Any) -> bool:
    """Function for comparing any two objects.

    If an equality test fails, assume not equal.

    Args:
        a (Any): Anything
        b (Any): Also anything

    Returns:
        bool: Are they the same
    """
    try:
        return cast(bool, a == b)
    except Exception:
        return False


class State(Generic[ST]):
    """The particular condition that something is in at a specific time.

    The states are implemented as
    `Descriptors <https://docs.python.org/3/howto/descriptor.html>`_
    which are associated to :class:`upstage.actor.Actor`.

    Note:
        The classes that use this descriptor must contain an ``env`` attribute.

    States are aware

    """

    def __init__(
        self,
        *,
        default: ST | None = None,
        frozen: bool = False,
        valid_types: type | tuple[type, ...] | None = None,
        recording: bool = False,
        record_duplicates: bool = False,
        default_factory: Callable[[], ST] | None = None,
        allow_none_default: bool = False,
    ) -> None:
        """Create a state descriptor for an Actor.

        The default can be set either with the value or the factory. Use the factory if
        the default needs to be a list, dict, or similar type of object. The default
        is used if both are present (not the factory).

        Setting frozen to True will throw an error if the value of the state is changed.

        The valid_types input will type-check when you initialize an actor.

        Recording enables logging the values of the state whenever they change, along
        with the simulation time. This attempts to deepcopy the value.

        When a state is a mutable type, such as a dictionary or Counter, state
        changes won't be recorded because the descriptor itself won't be modified
        through the __set__ call.

        Args:
            default (Any | None, optional): Default value of the state. Defaults to None.
            frozen (bool, optional): If the state is allowed to change. Defaults to False.
            valid_types (type | tuple[type, ...] | None, optional): Types allowed. Defaults to None.
            recording (bool, optional): If the state records itself. Defaults to False.
            record_duplicates (bool, optional): If the state records duplicate values.
                Defaults to False.
            default_factory (Callable[[], type] | None, optional): Default from function.
                Defaults to None.
            allow_none_default (bool, optional): Consider a `None` default to be
                valid
        """
        if default is None and default_factory is not None:
            default = default_factory()

        self._default = default
        self._frozen = frozen
        self._recording = recording
        self._record_duplicates = record_duplicates
        self._recording_callbacks: dict[Any, CALLBACK_FUNC] = {}
        self._allow_none_default = allow_none_default

        self._types: tuple[type, ...]

        if isinstance(valid_types, type):
            self._types = (valid_types,)
        elif valid_types is None:
            self._types = tuple()
        else:
            self._types = valid_types
        self.IGNORE_LOCK: bool = False

    def _do_record(self, instance: "Actor", value: ST, override: Any = None) -> None:
        """Record the value of the state.

        Args:
            instance (Actor): The actor holding the state
            value (ST): State value
            override (Any, optional): If given, record the override value
        """
        if not self._recording:
            return
        env = getattr(instance, "env", None)
        if env is None:
            raise SimulationError(
                f"Actor {instance} does not have an `env` attribute for state {self.name}"
            )
        use = value if override is None else override
        to_append = (env.now, deepcopy(use))
        if self.name not in instance._state_histories:
            instance._state_histories[self.name] = [to_append]
        elif self._record_duplicates or not _compare(
            to_append, instance._state_histories[self.name][-1]
        ):
            instance._state_histories[self.name].append(to_append)

    def _do_callback(self, instance: "Actor", value: ST) -> None:
        """Run callbacks for the state change.

        Args:
            instance (Actor): The actor holding the state
            value (Any): The value of the state
        """
        for _, callback in self._recording_callbacks.items():
            callback(instance, value)

    def _broadcast_change(self, instance: "Actor", name: str, value: ST) -> None:
        """Send state change values to nucleus.

        Args:
            instance (Actor): The actor holding the state
            name (str): The state's name
            value (Any): The state's value
        """
        # broadcast changes to the instance
        if instance._state_listener is not None:
            instance._state_listener.send_change(name, value)

    # NOTE: A dictionary as a descriptor doesn't work well,
    # because all the operations seem to happen *after* the get
    # NOTE: Lists also have the same issue that
    def __set__(self, instance: "Actor", value: ST) -> None:
        """Set the state's value.

        Args:
            instance (Actor): The actor holding the state
            value (Any): The state's value
        """
        if self._frozen:
            old_value = instance.__dict__.get(self.name, None)
            if old_value is not None:
                raise SimulationError(
                    f"State '{self}' on '{instance}' has already been frozen "
                    f"to value of {old_value}. It cannot be changed once set!"
                )

        if self._types and not isinstance(value, self._types):
            raise TypeError(f"{value} is of type {type(value)} not of type {self._types}")

        instance.__dict__[self.name] = value

        self._do_record(instance, value)
        self._do_callback(instance, value)

        self._broadcast_change(instance, self.name, value)

    def __get__(self, instance: "Actor", objtype: type | None = None) -> ST:
        if instance is None:
            # instance attribute accessed on class, return self
            return self  # pragma: no cover
        if self.name in instance._mimic_states:
            actor, name = instance._mimic_states[self.name]
            value = getattr(actor, name)
            self.__set__(instance, value)
        if self.name not in instance.__dict__:
            raise SimulationError(f"State {self.name} should have been set.")
        v = instance.__dict__[self.name]
        return cast(ST, v)

    def __set_name__(self, owner: "Actor", name: str) -> None:
        self.name = name

    def _set_default(self, instance: "Actor") -> None:
        """Set the state's value on the actor the default.

        For allowed None default, skip setting it. This will
        error on the get, which is expected.

        Args:
            instance (Actor): Actor holding the state.
        """
        if self._default is None:
            if self._allow_none_default:
                return
            raise SimulationError("State not allowed `None` default.")
        value = deepcopy(self._default)
        self.__set__(instance, value)

    def has_default(self) -> bool:
        """Check if a default exists.

        Returns:
            bool
        """
        if self._allow_none_default:
            return True
        return self._default is not None

    def _add_callback(self, source: Any, callback: CALLBACK_FUNC) -> None:
        """Add a recording callback.

        Args:
            source (Any): A key for the callback
            callback (Callable[[Actor, Any], None]): A function to call
        """
        self._recording_callbacks[source] = callback

    def _remove_callback(self, source: Any) -> None:
        """Remove a callback.

        Args:
            source (Any): The callback's key
        """
        del self._recording_callbacks[source]

    @property
    def is_recording(self) -> bool:
        """Check if the state is recording.

        Returns:
            bool
        """
        return self._recording


class DetectabilityState(State[bool]):
    """A state whose purpose is to indicate True or False.

    For consideration in the motion manager's <>LocationChangingState checks.
    """

    def __init__(self, *, default: bool = False, recording: bool = False) -> None:
        """Create the detectability state.

        Args:
            default (bool, optional): If the state starts on/off. Defaults to False.
            recording (bool, optional): If the state records. Defaults to False.
        """
        super().__init__(
            default=default,
            frozen=False,
            valid_types=(bool,),
            recording=recording,
        )

    def __set__(self, instance: "Actor", value: bool) -> None:
        """Set the detectability.

        Args:
            instance (Actor): The actor
            value (bool): The value to set
        """
        # Setting the default value shouldn't trigger the callback
        # to the motion manager.
        was_set = True
        if self.name not in instance.__dict__:
            was_set = False
        super().__set__(instance, value)
        if hasattr(instance.stage, "motion_manager") and was_set:
            mgr = instance.stage.motion_manager
            if not value:
                mgr._mover_not_detectable(instance)
            else:
                mgr._mover_became_detectable(instance)


class ActiveState(State, Generic[ST]):
    """Base class for states that change over time according to some rules.

    This class must be subclasses with an implemented `active` method.

    """

    def _active(self, instance: "Actor") -> Any:
        """Determine if the instance has an active state.

        Note:
            The instance must have two methods: ``get_active_state_data`` and
            ``_set_active_state_data``.

        Note:
            When you call ``activate_state`` from an actor, that is where
            you define the activity data. It is up to the Actor's subclass to
            make sure the activity data meet its needs.

            The first entry in the active data is always the time.
            Alternatively, you can call ``self.get_activity_data`` for some
            more data.

        """
        raise NotImplementedError("Method active not implemented.")

    def __get__(self, instance: "Actor", owner: type | None = None) -> ST:
        if instance is None:
            # instance attribute accessed on class, return self
            return self  # pragma: no cover
        if self.name not in instance.__dict__:
            # Just set the value to the default
            # Mutable types will be tricky here, so deepcopy them
            instance.__dict__[self.name] = deepcopy(self._default)  # pragma: no cover
        if self.name in instance._mimic_states:
            actor, name = instance._mimic_states[self.name]
            value = getattr(actor, name)
            self.__set__(instance, value)
            return cast(ST, value)
        # test if this instance is active or not
        res = self._active(instance)
        # comes back as None (not active), or if it can be obtained from dict
        if res is None:
            res = instance.__dict__[self.name]
        return cast(ST, res)

    def get_activity_data(self, instance: "Actor") -> dict[str, Any]:
        """Get the data useful for updating active states.

        Returns:
            dict[str, Any]: A dictionary with the state's pertinent data. Includes the actor's
                environment current time (``'now'``) and the value of the actor's
                state (``'state'``).

        """
        res = instance.get_active_state_data(self.name, without_update=True)
        res["now"] = instance.env.now
        res["value"] = instance.__dict__[self.name]
        return res

    def activate(self, instance: "Actor", task: Task | None = None) -> None:
        """Method to run when a state is activated.

        Used to help record the right data about the active state.

        Use this with __super__ for motion states to deactivate their motion from
        the motion manager.
        """
        self._do_record(instance, None, override=ActiveStatus.activating)

    def deactivate(self, instance: "Actor", task: Task | None = None) -> bool:
        """Optional method to override that is called when a state is deactivated.

        Useful for motion states to deactivate their motion from
        the motion manager.

        Defaults to any deactivation removing active state data.
        """
        # Returns if the state should be ignored
        # A False means the state is completely deactivated
        self._do_record(instance, None, override=ActiveStatus.deactivating)
        return False


class LinearChangingState(ActiveState[float]):
    """A state whose value changes linearly over time.

    When activating:

    >>> class Lin(Actor):
    >>>     x = LinearChangingState()
    >>>
    >>> def task(self, actor: Lin):
    >>>     actor.activate_state(
    >>>         name="x",
    >>>         task=self,
    >>>         rate=3.2,
    >>>     )
    """

    def _active(self, instance: "Actor") -> float | None:
        """Return a value to set based on time or some other criteria."""
        data = self.get_activity_data(instance)
        now: float = data["now"]
        current: float = data["value"]
        started: float | None = data.get("started_at", None)
        if started is None:
            # it's not currently active
            return None
        # The user needs to know what their active data looks like.
        # Alternatively, it could be defined in the state or the actor.
        rate: float = data["rate"]
        if now < started:
            raise SimulationError(
                f"Cannot set state '{self.name}' start time after now. "
                f"This probably happened because the active state was "
                f"set incorrectly."
            )
        value = (now - started) * rate
        return_value = current + value
        self.__set__(instance, return_value)
        instance._set_active_state_data(
            state_name=self.name,
            started_at=now,
            rate=rate,
        )
        return return_value


class CartesianLocationChangingState(ActiveState[CartesianLocation]):
    """A state that contains the location in 3-dimensional Cartesian space.

    Movement is along straight lines in that space.

    For activating:
        >>> actor.activate_state(
        >>>     state=<state name>,
        >>>     task=self, # usually
        >>>     speed=<speed>,
        >>>     waypoints=[
        >>>         List of CartesianLocation
        >>>     ]
        >>> )
    """

    def __init__(self, *, recording: bool = False):
        """Set a Location changing state.

        Defaults are disabled due to immutability of location objects.
        (We could copy it, but it seems like better practice to force inputting it at runtime.)

        Args:
            recording (bool, optional): Whether to record. Defaults to False.
        """
        super().__init__(
            default=None,
            frozen=False,
            default_factory=None,
            valid_types=(CartesianLocation,),
            recording=recording,
        )

    def _setup(self, instance: "Actor") -> None:
        """Initialize data about a path.

        Args:
            instance (Actor): The actor
        """
        data = self.get_activity_data(instance)
        current: CartesianLocation = data["value"]
        speed: float = data["speed"]
        waypoints: list[CartesianLocation] = data["waypoints"]
        # get the times, distances, and bearings from the waypoints
        times: list[float] = []
        distances: list[float] = []
        starts: list[CartesianLocation] = []
        vectors: list[list[float]] = []
        for wypt in waypoints:
            dist = wypt - current
            time = dist / speed
            times.append(time)
            distances.append(dist)
            starts.append(current.copy())
            vectors.append(_vector_subtract(wypt._as_array(), current._as_array()))
            current = wypt

        path_data = {
            "times": times,
            "distances": distances,
            "starts": starts,
            "vectors": vectors,
        }
        instance._set_active_state_data(
            self.name,
            started_at=data["now"],
            origin=data["value"],
            speed=speed,
            waypoints=waypoints,
            path_data=path_data,
        )
        # if there is a motion manager, notify it
        if hasattr(instance.stage, "motion_manager"):
            if not getattr(instance, "_is_rehearsing", False):
                instance.stage.motion_manager._start_mover(
                    instance,
                    speed,
                    [data["value"]] + waypoints,
                )

    def _get_index(self, path_data: dict[str, Any], time_elapsed: float) -> tuple[int, float]:
        """Find out how far along waypoints the state is.

        Args:
            path_data (dict[str, Any]): Data about the movement path
            time_elapsed (float): Time spent moving

        Returns:
            int: index in waypoints
            float: time spent on path
        """
        sum_t = 0.0
        t: float
        for i, t in enumerate(path_data["times"]):
            sum_t += t
            if time_elapsed <= (sum_t + 1e-12):
                return i, sum_t - t
        raise SimulationError(
            "CartesianLocation active state exceeded travel time: "
            f"elapsed: {time_elapsed}, maximum: {sum_t}"
        )

    def _get_remaining_waypoints(self, instance: "Actor") -> list[CartesianLocation]:
        """Convenience for getting waypoints left.

        Args:
            instance (Actor): The owning actor.

        Returns:
            list[CartesianLocation]: The waypoints left
        """
        data = self.get_activity_data(instance)
        current_time: float = data["now"]
        path_start_time: float = data["started_at"]
        elapsed = current_time - path_start_time
        idx, _ = self._get_index(data["path_data"], elapsed)
        return list(data["waypoints"][idx:])

    def _active(self, instance: "Actor") -> CartesianLocation | None:
        """Get the current value while active.

        Args:
            instance (Actor): The owning actor

        Returns:
            CartesianLocation | None: The current value
        """
        data = self.get_activity_data(instance)
        path_start_time: float | None = data.get("started_at", None)
        if path_start_time is None:
            # it's not active
            return None

        path_data: dict[str, Any] | None = data.get("path_data", None)
        if path_data is None:
            self._setup(instance)
            data = self.get_activity_data(instance)

        path_data: dict[str, Any] = data["path_data"]
        current_time: float = data["now"]
        elapsed = current_time - path_start_time
        if elapsed < 0:
            # Can probably only happen if active state is set incorrectly
            raise SimulationError(f"Cannot set state '{self.name}' start time in the future!")
        elif elapsed == 0:
            return_value: CartesianLocation = data["value"]  # pragma: no cover
        else:
            # Get the location along the waypoint path
            wypt_index, wypt_start = self._get_index(path_data, elapsed)
            time_along = elapsed - wypt_start
            path_time: float = path_data["times"][wypt_index]
            path_start: CartesianLocation = path_data["starts"][wypt_index]
            path_vector: list[float] = path_data["vectors"][wypt_index]
            time_frac = time_along / path_time
            direction_amount = [time_frac * v for v in path_vector]
            new_point = _vector_add(path_start._as_array(), direction_amount)

            # make the right kind of location object
            new_location = CartesianLocation(
                x=new_point[0],
                y=new_point[1],
                z=new_point[2],
            )
            return_value = new_location

            self.__set__(instance, return_value)

            # No new data needs to be added
            # Only the current time is needed once we run _setup()
            data["value"] = return_value
            instance._set_active_state_data(
                state_name=self.name,
                **data,
            )

        return return_value

    def deactivate(self, instance: "Actor", task: Task | None = None) -> bool:
        """Deactivate the motion.

        Args:
            instance (Actor): The owning actor
            task (Task): The task calling the deactivation.

        Returns:
            bool: _description_
        """
        if hasattr(instance.stage, "motion_manager"):
            if not getattr(instance, "_is_rehearsing", False):
                instance.stage.motion_manager._stop_mover(instance)
        return super().deactivate(instance, task)


class GeodeticLocationChangingState(ActiveState[GeodeticLocation]):
    """A state that contains a location around an ellipsoid that follows great-circle paths.

    Requires a distance model class that implements:
    1. distance_and_bearing
    2. point_from_bearing_dist
    and outputs objects with .lat and .lon attributes


    For activating:

    >>> actor.activate_state(
    >>>     state=<state name>,
    >>>     task=self, # usually
    >>>     speed=<speed>,
    >>>     waypoints=[
    >>>         List of CartesianLocation
    >>>     ]
    >>> )
    """

    def __init__(self, *, recording: bool = False) -> None:
        """Create the location changing state.

        Defaults are disabled due to immutability of location objects.
        (We could copy it, but it seems like better practice to force inputting it at runtime.)

        Args:
            recording (bool, optional): If the location is recorded. Defaults to False.
        """
        super().__init__(
            default=None,
            frozen=False,
            valid_types=(GeodeticLocation,),
            recording=recording,
        )

    def _setup(self, instance: "Actor") -> None:
        """Initialize data about a path."""
        STAGE = instance.stage
        data = self.get_activity_data(instance)
        current: GeodeticLocation = data["value"]
        speed: float = data["speed"]
        waypoints: list[GeodeticLocation] = data["waypoints"]
        # get the times, distances, and bearings from the waypoints
        times: list[float] = []
        distances: list[float] = []
        bearings: list[float] = []
        starts: list[GeodeticLocation] = []
        for wypt in waypoints:
            dist, bear = STAGE.stage_model.distance_and_bearing(
                (current.lat, current.lon),
                (wypt.lat, wypt.lon),
                units=STAGE.distance_units,
            )
            time = dist / speed
            times.append(time)
            distances.append(dist)
            bearings.append(bear)
            starts.append(current.copy())
            current = wypt

        path_data = {
            "times": times,
            "distances": distances,
            "bearings": bearings,
            "starts": starts,
        }
        instance._set_active_state_data(
            self.name,
            started_at=data["now"],
            origin=data["value"],
            speed=speed,
            waypoints=waypoints,
            path_data=path_data,
        )

        # if there is a motion manager, notify it
        if hasattr(STAGE, "motion_manager"):
            if not getattr(instance, "_is_rehearsing", False):
                STAGE.motion_manager._start_mover(
                    instance,
                    speed,
                    [data["value"]] + waypoints,
                )

    def _get_index(self, path_data: dict[str, Any], time_elapsed: float) -> tuple[int, float]:
        """Get the index of the waypoint the path is on.

        Args:
            path_data (dict[str, Any]): Data about the motion
            time_elapsed (float): Time spent on motion

        Returns:
            int: Index of the waypoint
            float: time elapsed
        """
        sum_t = 0.0
        t: float
        for i, t in enumerate(path_data["times"]):
            sum_t += t
            if time_elapsed <= (sum_t + 1e-4):  # near one second allowed
                return i, sum_t - t
        raise SimulationError(
            f"GeodeticLocation active state exceeded travel time: Elapsed: {time_elapsed}, "
            "Actual: {sum_t}"
        )

    def _get_remaining_waypoints(self, instance: "Actor") -> list[GeodeticLocation]:
        """Get waypoints left in travel.

        Args:
            instance (Actor): The owning actor.

        Returns:
            list[GeodeticLocation]: Waypoint remaining
        """
        data = self.get_activity_data(instance)
        current_time: float = data["now"]
        path_start_time: float = data["started_at"]
        elapsed = current_time - path_start_time
        idx, _ = self._get_index(data["path_data"], elapsed)
        wypts: list[GeodeticLocation] = data["waypoints"]
        return wypts[idx:]

    def _active(self, instance: "Actor") -> GeodeticLocation | None:
        """Get the value of the location while in motion.

        Args:
            instance (Actor): The owning actor.

        Returns:
            GeodeticLocation | None: Location while in motion. None if still.
        """
        STAGE = instance.stage
        data = self.get_activity_data(instance)
        path_start_time: float | None = data.get("started_at", None)
        if path_start_time is None:
            # it's not active
            return None

        path_data: dict[str, Any] | None = data.get("path_data", None)
        if path_data is None:
            self._setup(instance)
            data = self.get_activity_data(instance)

        path_data: dict[str, Any] = data["path_data"]

        current_time: float = data["now"]
        path_start_time: float = data["started_at"]
        elapsed = current_time - path_start_time

        if elapsed < 0:
            # Can probably only happen if active state is set incorrectly
            raise SimulationError(f"Cannot set state '{self.name}' start time in the future!")
        elif elapsed == 0:
            return_value: GeodeticLocation = data["value"]  # pragma: no cover
        else:
            # Get the location along the waypoint path
            wypt_index, wypt_start = self._get_index(path_data, elapsed)
            time_along = elapsed - wypt_start
            path_time: float = path_data["times"][wypt_index]
            path_dist: float = path_data["distances"][wypt_index]
            path_bearing: float = path_data["bearings"][wypt_index]
            path_start: GeodeticLocation = path_data["starts"][wypt_index]
            moved_distance = (time_along / path_time) * path_dist
            new_point = STAGE.stage_model.point_from_bearing_dist(
                (path_start.lat, path_start.lon),
                path_bearing,
                moved_distance,
                STAGE.distance_units,
            )
            # update the altitude
            waypoint: GeodeticLocation = data["waypoints"][wypt_index]
            alt_shift = waypoint.alt - path_start.alt
            alt_shift *= time_along / path_time
            new_alt = path_start.alt + alt_shift
            # make the right kind of location object
            lat, lon = new_point[0], new_point[1]
            new_location = GeodeticLocation(
                lat,
                lon,
                new_alt,
            )
            return_value = new_location

            self.__set__(instance, return_value)

            # No new data needs to be added
            # Only the current time is needed once we run _setup()
            instance._set_active_state_data(
                state_name=self.name,
                **data,
            )

        return return_value

    def deactivate(self, instance: "Actor", task: Task | None = None) -> bool:
        """Deactivate the state.

        Args:
            instance (Actor): The owning actor
            task (Task): The task doing the deactivating

        Returns:
            bool: If the state is all done
        """
        STAGE = instance.stage
        if hasattr(STAGE, "motion_manager"):
            if not getattr(instance, "_is_rehearsing", False):
                STAGE.motion_manager._stop_mover(instance)
        return super().deactivate(instance, task)


T = TypeVar("T", bound=Store | Container)


class ResourceState(State, Generic[T]):
    """A State class for States that are meant to be Stores or Containers.

    This should enable easier initialization of Actors with stores/containers or
    similar objects as states.

    No input is needed for the state if you define a default resource class in
    the class definition and do not wish to modify the default inputs of that
    class. You can also define default inputs for the resource instantiation.

    The input an Actor needs to receive for a ResourceState is a dictionary of:
    * 'kind': <class> (optional if you provided a default)
    * 'capacity': <numeric> (optional, works on stores and containers)
    * 'init': <numeric> (optional, works on containers)
    * key:value for any other input expected as a keyword argument by the resource class

    Note that the resource class given must accept the environment as the first
    positional argument. This is to maintain compatibility with simpy.

    Example:
        >>> class Warehouse(Actor):
        >>>     shelf = ResourceState[Store](default=Store)
        >>>     bucket = ResourceState[Container](
        >>>         default=Container,
        >>>         valid_types=(Container, SelfMonitoringContainer),
        >>>     )
        >>>     charger = ResourceState[Store](
        >>>         default=Store,
        >>>         default_kwargs={"capacity": 5},
        >>>     )
        >>>
        >>> wh = Warehouse(
        >>>     name='Depot',
        >>>     shelf={'capacity': 10},
        >>>     bucket={'kind': SelfMonitoringContainer, 'init': 30},
        >>> )
    """

    def __init__(
        self,
        *,
        default: Any | None = None,
        valid_types: type | tuple[type, ...] | None = None,
        default_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """Create a resource State decorator.

        Args:
            default (Any | None, optional): Default store/container class. Defaults to None.
            valid_types (type | tuple[type, ...] | None, optional): Valid store/container
                classes. Defaults to None.
            default_kwargs (dict[str, Any], optional): Kwargs to pass to the creation
                of the default store/container class.
        """
        if isinstance(valid_types, type):
            valid_types = (valid_types,)

        if valid_types:
            for v in valid_types:
                if not isinstance(v, type) or not issubclass(v, Store | Container):
                    raise UpstageError(f"Bad valid type for {self}: {v}")
        else:
            valid_types = (Store, Container)

        if default is not None and (
            not isinstance(default, type) or not issubclass(default, Store | Container)
        ):
            raise UpstageError(f"Bad default type for {self}: {default}")

        super().__init__(
            default=default,
            frozen=False,
            recording=False,
            valid_types=valid_types,
        )
        self._default_kwargs = default_kwargs.copy() if default_kwargs is not None else {}
        self._been_set: set[Actor] = set()

    def __set__(self, instance: "Actor", value: dict | Any) -> None:
        """Set the state value.

        Args:
            instance (Actor): The actor instance
            value (dict | Any): Either a dictionary of resource data OR an actual resource
        """
        if instance in self._been_set:
            raise UpstageError(
                f"State '{self}' on '{instance}' has already been created "
                "It cannot be changed once set!"
            )

        if not isinstance(value, dict):
            # we've been passed an actual resource, so save it and leave
            if not isinstance(value, self._types):
                raise UpstageError(f"Resource object: '{value}' is not an expected type.")
            instance.__dict__[self.name] = value
            self._been_set.add(instance)
            return

        resource_type = value.get("kind", self._default)
        if resource_type is None:
            raise UpstageError(f"No resource type (Store, e.g.) specified for {instance}")

        if self._types and not issubclass(resource_type, self._types):
            raise UpstageError(
                f"{resource_type} is of type {type(resource_type)} not of type {self._types}"
            )

        env = getattr(instance, "env", None)
        if env is None:
            raise UpstageError(
                f"Actor {instance} does not have an `env` attribute for state {self.name}"
            )
        kwargs = self._default_kwargs.copy()
        kwargs.update({k: v for k, v in value.items() if k != "kind"})
        try:
            resource_obj = resource_type(env, **kwargs)
        except TypeError as e:
            raise UpstageError(
                f"Bad argument input to resource state {self.name}"
                f" resource class {resource_type} :{e}"
            )
        except Exception as e:
            raise UpstageError(f"Exception in ResourceState init: {e}")

        instance.__dict__[self.name] = resource_obj
        self._been_set.add(instance)
        # remember what we did for cloning
        instance.__dict__["_memory_for_" + self.name] = kwargs.copy()

        self._broadcast_change(instance, self.name, value)

    def _set_default(self, instance: "Actor") -> None:
        """Set the default conditions.

        The empty dictionary input forces default to happen the right way.

        Args:
            instance (Actor): The actor holding this state.
        """
        self.__set__(instance, {})

    def __get__(self, instance: "Actor", owner: type | None = None) -> T:
        if instance is None:
            # instance attribute accessed on class, return self
            return self  # pragma: no cover
        if self.name not in instance.__dict__:
            self._set_default(instance)
        obj = instance.__dict__[self.name]
        if not issubclass(type(obj), Store | Container):
            raise UpstageError("Bad type of ResourceStatee")
        return cast(T, obj)

    def _make_clone(self, instance: "Actor", copy: T) -> T:
        """Method to support cloning a store or container.

        Args:
            instance (Actor): The owning actor
            copy (T): The store or container to copy

        Returns:
            T: The copied store or container
        """
        base_class = type(copy)
        memory: dict[str, Any] = instance.__dict__[f"_memory_for_{self.name}"]
        new = base_class(instance.env, **memory)  # type: ignore [arg-type]
        if isinstance(copy, Store) and isinstance(new, Store):
            new.items = list(copy.items)
        if isinstance(copy, Container) and isinstance(new, Container):
            # This is a particularity of simpy containers
            new._level = float(copy.level)
        return cast(T, new)


class CommunicationStore(ResourceState[Store]):
    """A State class for communications inputs.

    Used for automated finding of communication inputs on Actors by the CommsTransfer code.

    Follows the same rules for defaults as `ResourceState`, except this
    defaults to a SelfMonitoringStore without any user input.

    Only resources inheriting from simpy.Store will work for this state.
    Capacities are assumed infinite.

    The input an Actor needs to receive for a CommunicationStore is a dictionary of:
        >>> {
        >>>     'kind': <class> (optional)
        >>>     'mode': <string> (required)
        >>> }

    Example:
        >>> class Worker(Actor):
        >>>     walkie = CommunicationStore(mode="UHF")
        >>>     intercom = CommunicationStore(mode="loudspeaker")
        >>>
        >>> worker = Worker(
        >>>     name='Billy',
        >>>     walkie={'kind': SelfMonitoringStore},
        >>> )

    """

    def __init__(
        self,
        *,
        mode: str,
        default: type | None = None,
        valid_types: type | tuple[type, ...] | None = None,
    ):
        """Create a comms store.

        Args:
            mode (str): A mode to describe the comms channel.
            default (type | None, optional): Store class by default.
                Defaults to None.
            valid_types (type | tuple[type, ...] | None, optional): Valid store classes.
                Defaults to None.
        """
        if default is None:
            default = SelfMonitoringStore
        if valid_types is None:
            valid_types = (Store, SelfMonitoringStore)
        elif isinstance(valid_types, type):
            valid_types = (valid_types,)
        for v in valid_types:
            if not issubclass(v, Store):
                raise SimulationError("CommunicationStore must use a Store subclass")
        super().__init__(default=default, valid_types=valid_types)
        self._mode = mode



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""Data types for common operations. Currently just locations."""

from dataclasses import FrozenInstanceError
from math import degrees, radians, sqrt
from typing import Any

from upstage_des.base import UpstageBase
from upstage_des.math_utils import _vector_norm, _vector_subtract
from upstage_des.units import unit_convert

__all__ = (
    "CartesianLocation",
    "GeodeticLocation",
    "Location",
    "CartesianLocationData",
    "GeodeticLocationData",
)


class Location(UpstageBase):
    """An abstract class for representing a location in a space."""

    def copy(self) -> "Location":
        """Copy the location."""
        raise NotImplementedError("Subclass must implement copy.")

    @property
    def _repred_attrs(self) -> dict[str, Any]:
        """A way to customize how attributes are represented by __repr__.

        Returns:
            dict[str, Any]: Relevant attributes with name:value pairs.
        """
        return {
            k: v
            for k, v in self.__dict__.items()
            if not k.startswith("_") and k not in ["env", "stage"]
        }

    def _key(self) -> tuple[float, ...]:
        """A key used for hashing."""
        raise NotImplementedError("Location is intended to be subclassed.")

    def _to_tuple(self) -> tuple[float, ...]:
        """Return a tuple of the location.

        To be implemented in a subclass.

        Returns:
            tuple[float, ...]: Tuple of numbers describing a location.
        """
        raise NotImplementedError("Subclass must implement tuple.")

    def straight_line_distance(self, other: object) -> float:
        """Straight line distances b/w locations.

        Args:
            other (Location): Another location
        """
        raise NotImplementedError(
            "Subclass must implement a subtraction operator to calculate distance between Locations"
        )

    def __setattr__(self, name: str, value: Any) -> None:
        """Locations should be frozen, so this setattr restricts changing the values.

        Args:
            name (str): Attribute name
            value (Any): Attribute value
        """
        if hasattr(self, "_no_override") and name in self._no_override:
            raise FrozenInstanceError(f"Locations are disallowed from setting {name}")
        return super().__setattr__(name, value)

    def __sub__(self, other: object) -> float:
        """Subtract location.

        Args:
            other (Location): Another location
        """
        raise NotImplementedError(
            "Subclass must implement a subtraction operator to calculate distance between Locations"
        )

    def __eq__(self, other: object) -> bool:
        """Test for equality with another location.

        Args:
            other (object): The other location object

        Returns:
            bool: If it is equal.
        """
        raise NotImplementedError("Subclass must implement a equality comparison")

    def __hash__(self) -> int:
        raise NotImplementedError("Location is not intended for solo use.")

    def __repr__(self) -> str:
        """Customized the printable representation of the object.

        Does this by:
        1. Using all the dataclass fields that have `repr` set to True
        2. Representation for attributes can be customized by defining a
           `_get_repred_attrs` method that returns a dictionary of attributes
           keys and repr'ed values.
        """
        clean_attrs = [f"{key}={value}" for key, value in self._repred_attrs.items()]
        return f"{self.__class__.__name__}({', '.join(clean_attrs)})"


class CartesianLocation(Location):
    """A location that can be mapped to a 3-dimensional cartesian space."""

    def __init__(
        self,
        x: float,
        y: float,
        z: float = 0.0,
        *,
        use_altitude_units: bool = False,
    ) -> None:
        """A Cartesian (3D space) location.

        use_altitude_units, when false, means Z distance uses the "distance_units" unit.

        When true, it uses "altitude_units".

        Use UP.add_stage_variable("altitude_units", "ft"), e.g.

        Args:
            x (float): X dimension location
            y (float): Y dimension location
            z (float, optional): Z location. Defaults to 0.0.
            use_altitude_units (bool, optional): Use the sim's altitude units. Defaults to False.
        """
        super().__init__()
        self.x = x
        self.y = y
        self.z = z
        self.use_altitude_units = use_altitude_units
        self._no_override = ["x", "y", "z", "use_altitude_units"]

    @property
    def _repred_attrs(self) -> dict[str, str]:
        """Allows for attributes to be repr'ed based on the state of the units.

        Returns:
            dict[str, str]: Strings of attributes.
        """
        attrs = {}
        try:
            hor_units = self.stage.distance_units
        except AttributeError:
            hor_units = ""
        try:
            alt_units = (
                self.stage.altitude_units if self.use_altitude_units else self.stage.distance_units
            )
        except AttributeError:
            alt_units = ""
        for coordinate in ("x", "y"):
            attrs[coordinate] = f"{getattr(self, coordinate):,}{hor_units}"
        attrs["z"] = f"{self.z:,}{alt_units}"
        return attrs

    def _as_array(self) -> tuple[float, float, float]:
        """Make an array of consistent units for all dimensions.

        Returns:
            tuple[float, float, float]: A 1-D array of (x, y, z)
        """
        if self.use_altitude_units:
            height = unit_convert(self.z, self.stage.altitude_units, self.stage.distance_units)
        else:
            height = self.z
        return (self.x, self.y, height)

    def _to_tuple(self) -> tuple[float, float, float]:
        """Return a tuple of the location.

        Returns:
            tuple[float, float, float]: Latitude, longitude, altitude.
        """
        return self._as_array()

    def copy(self) -> "CartesianLocation":
        """Return a copy of the location.

        Returns:
            CartesianLocation
        """
        return self.__class__(
            x=self.x, y=self.y, z=self.z, use_altitude_units=self.use_altitude_units
        )

    def _key(self) -> tuple[float, float, float, bool]:
        """Key for hashing.

        Returns:
            tuple[float, float, float, bool]
        """
        return (self.x, self.y, self.z, self.use_altitude_units)

    def straight_line_distance(self, other: object) -> float:
        """Get the straight line distance between this and another location.

        Args:
            other (object): The other CartesianLocation point.
        """
        return self - other

    def __getitem__(self, idx: int) -> float:
        """Convenience way to get an xyz position by index.

        Args:
            idx (int): Index (xyz)

        Returns:
            float: Value at index
        """
        if not 0 <= idx <= 2:
            raise ValueError(f"CartesianLocation only has 3 indices (x, y, z), not a {idx}th index")
        return [self.x, self.y, self.z][idx]

    def __sub__(self, other: object) -> float:
        """Subtract one cartesian location from another.

        Distance is straight lint.

        Args:
            other (CartesianLocation): Another location

        Returns:
            float: Distance between this and another location.
        """
        if isinstance(other, CartesianLocation):
            sum_sq = sum((a - b) ** 2 for a, b in zip(self._as_array(), other._as_array()))
            return sqrt(sum_sq)
        else:
            raise ValueError(f"Cannot subtract {other.__class__.__name__} from a CartesianLocation")

    def __eq__(self, other: object) -> bool:
        """Test if two positions are the same.

        Uses a tolerance so this will be True for very close positions.

        Args:
            other (CartesianLocation): Another location

        Returns:
            bool: Is equal or not
        """
        if not isinstance(other, CartesianLocation):
            raise ValueError(f"Cannot compare {other.__class__.__name__} to a CartesianLocation")
        dist = self - other
        return bool(abs(dist) <= 0.00001)

    def __hash__(self) -> int:
        """Hash based on the key.

        Returns:
            int: The hash.
        """
        return hash(self._key())


class GeodeticLocation(Location):
    """A Location that can be mapped to the surface of a spheroid (a.k.a. ellipsoid).

    More specifically, a Location representing somewhere on an ellipsoid, with Latitude,
    Longitude, and Altitude that uses the geodetic datum (or geodetic system).  Can be
    used to define a location on Earth or other planetary bodies.

    Units for the horizontal datum (i.e., `lat` and `lon`) can be Decimal Degrees or
    Radians, depending on the value of `units`, the vertical datum (i.e., `alt`) is
    assumed to be in meters.

    Subtraction represents a great circle distance, NOT a true 3D straight-line distance.

    Speeds used for this location type will represent ground speed, which allows
    the class to ignore solving the exact altitude change in the path.

    Units are an input to the location type.

    The ellipsoid model must have a `.distance(Location1, Location2)` method
    that looks for .lat and .lon.

    `altitude` is 0.0 by default.

    `lat` (latitude) and `lon` (longitude) are in degrees by default.
        if using radians, set `in_radians` to `True`.

    """

    def __init__(
        self,
        lat: float,
        lon: float,
        alt: float = 0.0,
        *,
        in_radians: bool = False,
    ) -> None:
        """A location on a geodetic (Earth).

        Altitude uses the "altitude_units" stage variable.

        Args:
            lat (float): Latitude (North/South)
            lon (float): Longitude (East/West)
            alt (float, optional): Altitude. Defaults to 0.0.
            in_radians (bool, optional): If the lat/lon are in radians or degrees.
                Defaults to False.

        Returns:
            GeodeticLocation
        """
        super().__init__()
        self.lat = lat
        self.lon = lon
        self.alt = alt
        self.in_radians = in_radians
        self._no_override = ["lat", "lon", "alt", "in_radians"]

    @property
    def _repred_attrs(self) -> dict[str, str]:
        """Allows for attributes to be repr'ed based on the state of the units.

        Returns:
            dict[str, str]: Strings of attributes.
        """
        attrs = {}
        units = "rad" if self.in_radians else "°"
        try:
            alt_units = self.stage.altitude_units
        except AttributeError:
            alt_units = ""
        for coordinate in ("lat", "lon"):
            attrs[coordinate] = f"{getattr(self, coordinate)}{units}"
        attrs["alt"] = f"{self.alt:,}{alt_units}"
        return attrs

    def _to_tuple(self) -> tuple[float, float, float]:
        """Return a tuple of the location.

        Returns:
            tuple[float, float, float]: Latitude, longitude, altitude.
        """
        return (self.lat, self.lon, self.alt)

    def latlon(self) -> tuple[float, float]:
        """Return a tuple of just lat/lon as degrees.

        Returns:
            tuple[float, float]: Latitude and longitude in degrees.
        """
        s = self.to_degrees()
        return (s.lat, s.lon)

    def to_radians(self) -> "GeodeticLocation":
        """Convert to radians, if already in radians, return self.

        Returns:
            GeodeticLocation
        """
        if self.in_radians:
            return self
        kwargs: dict[str, float | bool] = {"alt": self.alt}
        for coordinate in ("lat", "lon"):
            kwargs[coordinate] = radians(getattr(self, coordinate))
        kwargs["in_radians"] = True
        return self.__class__(**kwargs)  # type: ignore [arg-type]

    def to_degrees(self) -> "GeodeticLocation":
        """Convert to degrees, if already in degrees, return self.

        Returns:
            GeodeticLocation
        """
        if not self.in_radians:
            return self
        kwargs: dict[str, float | bool] = {"alt": self.alt}
        for coordinate in ("lat", "lon"):
            kwargs[coordinate] = degrees(getattr(self, coordinate))
        kwargs["in_radians"] = False
        return self.__class__(**kwargs)  # type: ignore [arg-type]

    def _key(self) -> tuple[float, float, float, bool]:
        """A key for hashing.

        Returns:
            tuple[float, float, float, bool]: All values.
        """
        return (self.lat, self.lon, self.alt, self.in_radians)

    def copy(self) -> "GeodeticLocation":
        """Copy the location.

        Returns:
            GeodeticLocation
        """
        return self.__class__(
            lat=self.lat,
            lon=self.lon,
            alt=self.alt,
            in_radians=self.in_radians,
        )

    def dist_with_altitude(self, other: "GeodeticLocation") -> float:
        """Get the distance between two points with an altitude component.

        Args:
            other (GeodeticLocation): The other point

        Returns:
            float: Distance - pythagorean of great-circle and altitude
        """
        dist = self - other
        alt = abs(self.alt - other.alt)
        alt = unit_convert(alt, self.stage.altitude_units, self.stage.distance_units)
        full_dist: float = sqrt(alt**2 + dist**2)
        return full_dist

    def straight_line_distance(self, other: object) -> float:
        """Straight-line distance, using ECEF.

        This won't account for horizon.

        Args:
            other (GeodeticLocation): The other point

        Returns:
            float: Distance
        """
        if not isinstance(other, GeodeticLocation):
            raise TypeError(f"Cannot subtract a {other.__class__.__name__} from a GeodeticLocation")
        lat, lon, alt = self.to_degrees()._to_tuple()
        alt = unit_convert(alt, self.stage.altitude_units, "m")
        ecef_self = self.stage.stage_model.lla2ecef([(lat, lon, alt)])[0]

        lat, lon, alt = other.to_degrees()._to_tuple()
        alt = unit_convert(alt, self.stage.altitude_units, "m")
        ecef_other = self.stage.stage_model.lla2ecef([(lat, lon, alt)])[0]

        dist_meters = float(_vector_norm(_vector_subtract(ecef_other, ecef_self)))
        dist_units = unit_convert(dist_meters, "m", self.stage.distance_units)
        return dist_units

    def __getitem__(self, idx: int) -> float:
        """Convenience way to get an xyz position by index.

        Args:
            idx (int): Index (lat/lon/alt)

        Returns:
            float: Value at index
        """
        if not 0 <= idx <= 2:
            raise IndexError(
                f"GeodeticLocation only has 3 indices (lat, lon, alt), not a {idx}th index"
            )
        return [self.lat, self.lon, self.alt][idx]

    def __sub__(self, other: object) -> float:
        """Find the great circle distance between Geodetic points.

        Args:
            other (GeodeticLocation): Another location

        Returns:
            float: distance in stage "distance_units" units.
        """
        if not isinstance(other, GeodeticLocation):
            raise ValueError(
                f"Cannot subtract a {other.__class__.__name__} from a GeodeticLocation"
            )
        # distances presume positions are in degrees
        dlat, dlon = self.to_degrees()._key()[:2]
        olat, olon = other.to_degrees()._key()[:2]
        dist = self.stage.stage_model.distance(
            (dlat, dlon),
            (olat, olon),
            units=self.stage.distance_units,
        )
        return dist

    def __eq__(self, other: object) -> bool:
        """Test if two locations are the same.

        No tolerance is applied here.

        Args:
            other (GeodeticLocation): Another location

        Returns:
            bool: Close enough or not.
        """
        if not isinstance(other, GeodeticLocation):
            raise ValueError(f"Cannot compare a {other.__class__.__name__} to a GeodeticLocation")
        if other.in_radians != self.in_radians:
            if self.in_radians:
                other = other.to_radians()
            else:
                other = other.to_degrees()
        return all(
            getattr(self, dimension) == getattr(other, dimension)
            for dimension in ["lat", "lon", "alt"]
        )

    def __hash__(self) -> int:
        """Hash based on the key.

        Returns:
            int: The hash.
        """
        return hash(self._key())


class CartesianLocationData:
    """Object for storing caretesian data without an environment."""

    def __init__(
        self,
        x: float,
        y: float,
        z: float = 0.0,
        *,
        use_altitude_units: bool = True,
    ) -> None:
        """Cartesian data storage.

        Args:
            x (float): _description_
            y (float): _description_
            z (float, optional): _description_. Defaults to 0.0.
            use_altitude_units (bool, optional): _description_. Defaults to True.
        """
        self.x = x
        self.y = y
        self.z = z
        self.use_altitude_unts = use_altitude_units

    def make_location(self) -> CartesianLocation:
        """Create a location from the data.

        Do this inside an environment contex.t

        Returns:
            CartesianLocation: The location object.
        """
        return CartesianLocation(
            x=self.x, y=self.y, z=self.z, use_altitude_units=self.use_altitude_unts
        )

    def __eq__(self, value: Any) -> bool:
        """Test for equality of two cartesian locations data objects."""
        if not isinstance(value, CartesianLocationData):
            raise ValueError(
                f"Cannot compare a {value.__class__.__name__} to a CartesianLocationData"
            )
        check = ["x", "y", "z"]
        return all(getattr(self, c) == getattr(value, c) for c in check)


class GeodeticLocationData:
    """Object for storing geodetic data without an environment."""

    def __init__(
        self,
        lat: float,
        lon: float,
        alt: float = 0.0,
        *,
        in_radians: bool = False,
    ) -> None:
        """Geodetic data storage.

        Args:
            lat (float): _description_
            lon (float): _description_
            alt (float, optional): _description_. Defaults to 0.0.
            in_radians (bool, optional): _description_. Defaults to False.
        """
        self.lat = lat
        self.lon = lon
        self.alt = alt
        self.in_radians = in_radians

    def make_location(self) -> GeodeticLocation:
        """Create a location from the data.

        Do this inside an environment contex.t

        Returns:
            GeodeticLocation: The location object.
        """
        return GeodeticLocation(
            lat=self.lat,
            lon=self.lon,
            alt=self.alt,
            in_radians=self.in_radians,
        )

    def __eq__(self, value: Any) -> bool:
        """Test for equality of two cartesian locations data objects."""
        if not isinstance(value, GeodeticLocationData):
            raise ValueError(
                f"Cannot compare a {value.__class__.__name__} to a GeodeticLocationData"
            )
        other = [value.lat, value.lon]
        if value.in_radians != self.in_radians:
            if self.in_radians:
                other = list(map(radians, other))
            else:
                other = list(map(degrees, other))
        angles = [self.lat, self.lon] == other
        return angles and self.alt == value.alt



"""Test the data recording/reporting capabilities."""

from collections import Counter
from dataclasses import dataclass

import simpy as SIM

import upstage_des.api as UP
from upstage_des.data_utils import create_location_table, create_table
from upstage_des.type_help import SIMPY_GEN


@dataclass
class Information:
    value_1: int
    value_2: float


class Cashier(UP.Actor):
    items_scanned = UP.State[int](recording=True)
    other = UP.State[float]()
    cue = UP.State[UP.SelfMonitoringStore]()
    cue2 = UP.ResourceState[UP.SelfMonitoringContainer](default=UP.SelfMonitoringContainer)
    time_working = UP.LinearChangingState(default=0.0, recording=True, record_duplicates=True)
    info = UP.State[Information](recording=True)


class Cart(UP.Actor):
    location = UP.CartesianLocationChangingState(recording=True)
    location_two = UP.CartesianLocationChangingState(recording=True)
    holding = UP.State[float](default=0.0, recording=True)
    some_data = UP.State[dict[str, float]](recording=True)


def test_data_reporting() -> None:
    with UP.EnvironmentContext() as env:
        t = UP.Task()
        cash = Cashier(
            name="Ertha",
            other=0.0,
            items_scanned=0,
            cue=UP.SelfMonitoringStore(env),
            info=Information(0, 0),
        )

        cash2 = Cashier(
            name="Bertha",
            other=0.0,
            items_scanned=0,
            cue=UP.SelfMonitoringStore(env),
            info=Information(1, 1),
        )
        store = UP.SelfMonitoringFilterStore(env, name="Store Test")
        cart = Cart(
            name="Wobbly Wheel",
            location=UP.CartesianLocation(1.0, 1.0),
            location_two=UP.CartesianLocation(1.0, 1.0),
            some_data={"exam": 2.0},
        )

        for c in [cash, cash2]:
            c.items_scanned += 1
            c.cue.put("A")
            c.cue2.put(10)
            c.time_working = 0.0
        cash.other = 3.0

        cart.activate_location_state(
            state="location",
            speed=2.0,
            waypoints=[UP.CartesianLocation(7.0, 6.0)],
            task=t,
        )
        cart.activate_location_state(
            state="location_two",
            speed=2.0,
            waypoints=[UP.CartesianLocation(-7.0, -6.0)],
            task=t,
        )

        env.run(until=0.1)
        for c in [cash, cash2]:
            c.activate_linear_state(
                state="time_working",
                rate=1.0,
                task=t,
            )

        env.run(until=1)
        cart.location
        cart.location_two
        cash.items_scanned += 2
        store.put("XYZ")
        cash.info.value_1 = 3
        cash.record_state("info")
        cash2.info.value_2 = 4.3
        cash2.record_state("info")

        for c in [cash, cash2]:
            c.cue.put("B")
            c.cue2.put(3)
            c.time_working

        env.run(until=2)
        cart.location
        cash.items_scanned += 1
        env.run(until=3)
        cart.location
        cart.location_two
        cart.some_data["exam"] = 4.0
        cart.record_state("some_data")

        cart.deactivate_state(state="location", task=t)
        cart.deactivate_state(state="location_two", task=t)

        cash2.deactivate_state(state="time_working", task=t)

        for c in [cash, cash2]:
            c.cue.get()
            c.cue2.get(2)
            c.time_working

        cash.items_scanned = -1
        env.run(until=3.3)
        cart.location
        cart.location_two = UP.CartesianLocation(-1.0, -1.0)
        cart.some_data["new"] = 123.45
        cart.record_state("some_data")
        store.put("ABC")
        env.run()
        cart.location
        cart.location_two
        for c in [cash, cash2]:
            c.time_working

        state_table, cols = create_table()
        all_state_table, all_cols = create_table(skip_locations=False)
        loc_state_table, loc_cols = create_location_table()
        new_table, _ = create_table(skip_locations=True, save_static=True)
        orig_table, _ = create_table(skip_locations=True, save_static=False)

    ctr = Counter([row[:3] for row in state_table])
    assert ctr[("Ertha", "Cashier", "items_scanned")] == 5
    assert ctr[("Ertha", "Cashier", "cue")] == 4
    assert ctr[("Ertha", "Cashier", "cue2")] == 4
    assert ctr[("Ertha", "Cashier", "time_working")] == 6
    assert ctr[("Bertha", "Cashier", "items_scanned")] == 2
    assert ctr[("Bertha", "Cashier", "cue")] == 4
    assert ctr[("Bertha", "Cashier", "cue2")] == 4
    assert ctr[("Bertha", "Cashier", "time_working")] == 5
    assert ctr[("Store Test", "SelfMonitoringFilterStore", "Resource")] == 3
    assert ctr[("Ertha", "Cashier", "info.value_1")] == 2
    assert ctr[("Ertha", "Cashier", "info.value_2")] == 2
    assert ctr[("Bertha", "Cashier", "info.value_1")] == 2
    assert ctr[("Bertha", "Cashier", "info.value_2")] == 2
    # Test for default values untouched in the sim showing up in the data.
    assert ctr[("Wobbly Wheel", "Cart", "holding")] == 1
    assert ctr[("Wobbly Wheel", "Cart", "some_data.exam")] == 3
    assert ctr[("Wobbly Wheel", "Cart", "some_data.new")] == 1
    row = [r for r in state_table if r[:3] == ("Wobbly Wheel", "Cart", "holding")][0]
    assert row[4] == 0
    assert row[3] == 0.0
    # Continuing as before
    assert len(state_table) == 50
    assert cols == all_cols
    assert cols == [
        "Entity Name",
        "Entity Type",
        "State Name",
        "Time",
        "Value",
        "Activation Status",
    ]

    ctr = Counter([row[:3] for row in all_state_table])
    assert ctr[("Ertha", "Cashier", "items_scanned")] == 5
    assert ctr[("Ertha", "Cashier", "cue")] == 4
    assert ctr[("Ertha", "Cashier", "cue2")] == 4
    assert ctr[("Ertha", "Cashier", "time_working")] == 6
    assert ctr[("Bertha", "Cashier", "items_scanned")] == 2
    assert ctr[("Bertha", "Cashier", "cue")] == 4
    assert ctr[("Bertha", "Cashier", "cue2")] == 4
    assert ctr[("Bertha", "Cashier", "time_working")] == 5
    assert ctr[("Store Test", "SelfMonitoringFilterStore", "Resource")] == 3
    assert ctr[("Wobbly Wheel", "Cart", "holding")] == 1
    assert ctr[("Wobbly Wheel", "Cart", "location")] == 4
    assert ctr[("Wobbly Wheel", "Cart", "location_two")] == 4
    assert len(all_state_table) == 38 + 8 + 12

    assert loc_cols == [
        "Entity Name",
        "Entity Type",
        "State Name",
        "Time",
        "X",
        "Y",
        "Z",
        "Activation Status",
    ]
    assert len(loc_state_table) == 8
    assert loc_state_table[-1] == (
        "Wobbly Wheel",
        "Cart",
        "location_two",
        3.3,
        -1.0,
        -1.0,
        0.0,
        "inactive",
    )

    match1 = ("Ertha", "Cashier", "other", 0.0, 3.0, "Last Seen")
    assert match1 in new_table
    assert match1 not in all_state_table

    match2 = ("Bertha", "Cashier", "other", 0.0, 0.0, "Last Seen")
    assert match2 in new_table
    assert match2 not in all_state_table

    # Only the two "other" states should show up
    assert len(new_table) - len(orig_table) == 2


def test_store_failure() -> None:
    class Exam(UP.Actor):
        a_store = UP.ResourceState[SIM.Store](default=SIM.Store)

    with UP.EnvironmentContext() as env:
        ex = Exam(name="example")

        def _proc() -> SIMPY_GEN:
            yield ex.a_store.put("a thing")
            yield env.timeout(1.0)
            assert ex.a_store.items == ["a thing"]
            yield ex.a_store.get()

        env.process(_proc())
        env.run()
        assert env.now == 1
        assert ex.a_store.items == []

        data, cols = create_table()
        assert data == []


if __name__ == "__main__":
    test_store_failure()



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from collections.abc import Generator

import simpy as SIM

import upstage_des.api as UP
from upstage_des.type_help import TASK_GEN


class Cashier(UP.Actor):
    scan_speed = UP.State[float](
        valid_types=(float,),
        frozen=True,
    )
    time_until_break = UP.State[float](
        default=120.0,
        valid_types=(float,),
        frozen=True,
    )
    breaks_until_done = UP.State[int](default=2, valid_types=int)
    breaks_taken = UP.State[int](default=0, valid_types=int, recording=True)
    items_scanned = UP.State[int](
        default=0,
        valid_types=(int,),
        recording=True,
    )
    time_scanning = UP.LinearChangingState(
        default=0.0,
        valid_types=(float,),
    )


class CheckoutLane(UP.Actor):
    customer_queue = UP.ResourceState[UP.SelfMonitoringStore](
        default=UP.SelfMonitoringStore,
    )


class StoreBoss(UP.UpstageBase):
    def __init__(self, lanes: list[CheckoutLane]) -> None:
        self.lanes = lanes
        self._lane_map: dict[CheckoutLane, Cashier] = {}

    def get_lane(self, cashier: Cashier) -> CheckoutLane:
        possible = [lane for lane in self.lanes if lane not in self._lane_map]
        lane = self.stage.random.choice(possible)
        self._lane_map[lane] = cashier
        return lane


class GoToWork(UP.Task):
    def task(self, *, actor: Cashier) -> TASK_GEN:
        """Go to work"""
        yield UP.Wait(15.0)


class TalkToBoss(UP.DecisionTask):
    def make_decision(self, *, actor: Cashier) -> None:
        """Zero-time task to get information."""
        boss: StoreBoss = self.stage.boss
        lane = boss.get_lane(actor)
        self.set_actor_knowledge(actor, "checkout_lane", lane, overwrite=False)
        actor.breaks_taken = 0
        self.set_actor_knowledge(actor, "start_time", self.env.now)


class WaitInLane(UP.Task):
    def task(self, *, actor: Cashier) -> TASK_GEN:
        """Wait until break time, or a customer."""
        lane: CheckoutLane = self.get_actor_knowledge(
            actor,
            "checkout_lane",
            must_exist=True,
        )
        customer_arrival = UP.Get(lane.customer_queue)

        start_time = self.get_actor_knowledge(
            actor,
            "start_time",
            must_exist=True,
        )
        break_start = start_time + actor.time_until_break
        wait_until_break = break_start - self.env.now
        if wait_until_break < 0:
            self.set_actor_task_queue(actor, ["Break"])
            return

        break_event = UP.Wait(wait_until_break)

        yield UP.Any(customer_arrival, break_event)

        if customer_arrival.is_complete():
            customer: int = customer_arrival.get_value()
            self.set_actor_knowledge(actor, "customer", customer, overwrite=True)
        else:
            customer_arrival.cancel()
            self.set_actor_task_queue(actor, ["Break"])


class DoCheckout(UP.Task):
    def task(self, *, actor: Cashier) -> TASK_GEN:
        """Do the checkout"""
        items: int = self.get_actor_knowledge(
            actor,
            "customer",
            must_exist=True,
        )
        per_item_time = actor.scan_speed / items
        actor.activate_linear_state(
            state="time_scanning",
            rate=1.0,
            task=self,
        )
        for _ in range(items):
            yield UP.Wait(per_item_time)
            actor.items_scanned += 1
        actor.deactivate_all_states(task=self)
        # assume 2 minutes to take payment
        yield UP.Wait(2.0)


class Break(UP.DecisionTask):
    def make_decision(self, *, actor: Cashier) -> None:
        """Decide what kind of break we are taking."""
        actor.breaks_taken += 1
        if actor.breaks_taken == actor.breaks_until_done:
            self.set_actor_task_queue(actor, ["NightBreak"])
        elif actor.breaks_taken > actor.breaks_until_done:
            raise UP.SimulationError("Too many breaks taken")
        else:
            self.set_actor_task_queue(actor, ["ShortBreak"])


class ShortBreak(UP.Task):
    def task(self, *, actor: Cashier) -> TASK_GEN:
        """Take a short break."""
        yield UP.Wait(15.0)
        self.set_actor_knowledge(actor, "start_time", self.env.now, overwrite=True)


class NightBreak(UP.Task):
    def task(self, *, actor: Cashier) -> TASK_GEN:
        """Go home and rest."""
        self.clear_actor_knowledge(actor, "checkout_lane")
        yield UP.Wait(60 * 12.0)


task_classes = {
    "GoToWork": GoToWork,
    "TalkToBoss": TalkToBoss,
    "WaitInLane": WaitInLane,
    "DoCheckout": DoCheckout,
    "Break": Break,
    "ShortBreak": ShortBreak,
    "NightBreak": NightBreak,
}

task_links = {
    "GoToWork": UP.TaskLinks(default="TalkToBoss", allowed=["TalkToBoss"]),
    "TalkToBoss": UP.TaskLinks(default="WaitInLane", allowed=["WaitInLane"]),
    "WaitInLane": UP.TaskLinks(default="DoCheckout", allowed=["DoCheckout", "Break"]),
    "DoCheckout": UP.TaskLinks(default="WaitInLane", allowed=["WaitInLane"]),
    "Break": UP.TaskLinks(default="ShortBreak", allowed=["ShortBreak", "NightBreak"]),
    "ShortBreak": UP.TaskLinks(default="WaitInLane", allowed=["WaitInLane"]),
    "NightBreak": UP.TaskLinks(default="GoToWork", allowed=["GoToWork"]),
}


cashier_task_network = UP.TaskNetworkFactory(
    name="CashierJob",
    task_classes=task_classes,
    task_links=task_links,
)


def customer_spawner(
    env: SIM.Environment,
    lanes: list[CheckoutLane],
) -> Generator[SIM.Event, None, None]:
    # sneaky way to get access to stage
    stage = lanes[0].stage
    while True:
        hrs = env.now / 60
        time_of_day = hrs // 24
        if time_of_day <= 8 or time_of_day >= 15.5:
            time_until_open = (24 - time_of_day) + 8
            yield env.timeout(time_until_open)

        lane_pick = stage.random.choice(lanes)
        number_pick = stage.random.randint(3, 17)
        yield lane_pick.customer_queue.put(number_pick)
        yield UP.Wait.from_random_uniform(5.0, 30.0).as_event()


def test_cashier_example() -> None:
    with UP.EnvironmentContext(initial_time=8 * 60) as env:
        UP.add_stage_variable("time_unit", "min")
        cashier = Cashier(
            name="Bob",
            scan_speed=1.0,
            time_until_break=120.0,
            breaks_until_done=4,
            debug_log=True,
        )
        lane_1 = CheckoutLane(name="Lane 1")
        lane_2 = CheckoutLane(name="Lane 2")
        boss = StoreBoss(lanes=[lane_1, lane_2])

        UP.add_stage_variable("boss", boss)

        net = cashier_task_network.make_network()
        cashier.add_task_network(net)
        cashier.start_network_loop(net.name, "GoToWork")

        customer_proc = customer_spawner(env, [lane_1, lane_2])
        _ = env.process(customer_proc)

        env.run(until=20 * 60)

    for line in cashier.get_log():
        print(line)


if __name__ == "__main__":
    test_cashier_example()



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""Declare the UPSTAGE version."""

__authors__ = "UPSTAGE Contributors, GTRI"
__version__ = "0.3.2"



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""Stores that monitor/record their items over time."""

from collections.abc import Callable
from typing import Any

from simpy import Container, Environment, Event, FilterStore, Store
from simpy.resources.container import ContainerGet, ContainerPut
from simpy.resources.store import FilterStoreGet, StoreGet, StorePut

from upstage_des.base import SPECIAL_ENTITY_CONTEXT_VAR, NamedUpstageEntity

from .container import ContinuousContainer
from .reserve import ReserveContainer
from .sorted import SortedFilterStore, _SortedFilterStoreGet

__all__ = (
    "SelfMonitoringStore",
    "SelfMonitoringFilterStore",
    "SelfMonitoringContainer",
    "SelfMonitoringContinuousContainer",
    "SelfMonitoringSortedFilterStore",
    "SelfMonitoringReserveContainer",
    "MonitoringMixin",
)

RECORDER_FUNC = Callable[[list[Any]], int]
MONITORING_ENTITY_GROUP = "monitored"


class MonitoringMixin(NamedUpstageEntity, skip_classname=True):
    """Base class for matching Monitored types."""

    name: str | None
    _quantities: list[tuple[float, Any]]

    def _add_special_group(self) -> None:
        """Add self the the monitored context group.

        Called by the NamedUpstageEntity on group inits.
        """
        ans = SPECIAL_ENTITY_CONTEXT_VAR.get().monitored
        if self in ans:
            return
        ans.append(self)


class SelfMonitoringStore(
    Store,
    MonitoringMixin,
    skip_classname=True,
):
    """A self-monitoring version of the SimPy Store."""

    def __init__(
        self,
        env: Environment,
        capacity: float | int = float("inf"),
        item_func: RECORDER_FUNC | None = None,
        name: str | None = None,
    ) -> None:
        """A monitoring version of the SimPy Store.

        Since Stores contain a list of objects, recording the mutable list is replaced
        by recording the length of the list by default. Otherwise, input the item_func
        to take the store items and return a recordable value.

        Args:
            env (Environment): SimpyEnvironment.
            capacity (float | int, optional): Capacity of the store. Defaults to float("inf").
            item_func (RECORDER_FUNC | None, optional): Function to create recorded values.
                Defaults to None.
            name (str, optional): The name of the store, if it doesn't exist as a state.
                Defaults to None.
        """
        super().__init__(env, capacity=capacity)
        self.name = name
        self.item_func = item_func if item_func is not None else len
        self._quantities = [(self._env.now, self.item_func(self.items))]

    def _record(self, call: str) -> None:
        v = self.item_func(self.items)
        if v != self._quantities[-1][1] or call == "environment":
            self._quantities.append((self._env.now, v))

    def _trigger_put(self, event: Event) -> None:  # type: ignore [override]
        super()._trigger_put(event)
        self._record("put")

    def _trigger_get(self, event: Event) -> None:  # type: ignore [override]
        super()._trigger_get(event)
        self._record("get")

    def _do_put(self, event: StorePut) -> None:
        super()._do_put(event)
        if event.triggered:
            self._record("put")
        return None

    def _do_get(self, event: StoreGet) -> None:
        super()._do_get(event)
        if event.triggered:
            self._record("get")
        return None


class SelfMonitoringFilterStore(
    FilterStore,
    MonitoringMixin,
    skip_classname=True,
):
    """A self-monitoring version of the SimPy FilterStore."""

    def __init__(
        self,
        env: Environment,
        capacity: float | int = float("inf"),
        item_func: RECORDER_FUNC | None = None,
        name: str | None = None,
    ) -> None:
        """A monitoring version of the SimPy FilterStore.

        Since Stores contain a list of objects, recording the mutable list is replaced
        by recording the length of the list by default. Otherwise, input the item_func
        to take the store items and return a recordable value.

        Args:
            env (Environment): SimpyEnvironment.
            capacity (float | int, optional): Capacity of the store. Defaults to float("inf").
            item_func (RECORDER_FUNC | None, optional): Function to create recorded values.
                Defaults to None.
            name (str, optional): The name of the store, if it doesn't exist as a state.
                Defaults to None.
        """
        super().__init__(env, capacity=capacity)
        self.name = name
        self.item_func = item_func if item_func is not None else len
        self._quantities = [(self._env.now, self.item_func(self.items))]

    def _record(self, call: str) -> None:
        v = self.item_func(self.items)
        if v != self._quantities[-1][1] or call == "environment":
            self._quantities.append((self._env.now, v))

    def _trigger_put(self, event: Event) -> None:  # type: ignore [override]
        super()._trigger_put(event)
        self._record("put")

    def _trigger_get(self, event: Event) -> None:  # type: ignore [override]
        super()._trigger_get(event)
        self._record("get")

    def _do_put(self, event: StorePut) -> None:
        super()._do_put(event)
        if event.triggered:
            self._record("put")

    def _do_get(self, event: FilterStoreGet) -> bool | None:  # type: ignore [override]
        ans = super()._do_get(event)
        if event.triggered:
            self._record("get")
        return ans


class SelfMonitoringContainer(
    Container,
    MonitoringMixin,
    skip_classname=True,
):
    """A self-monitoring version of the SimPy Container."""

    def __init__(
        self,
        env: Environment,
        capacity: float = float("inf"),
        init: float = 0.0,
        name: str | None = None,
    ) -> None:
        """A monitoring version of a SimPy container.

        Args:
            env (Environment): SimPy environment.
            capacity (float, optional): Capacity of the container. Defaults to float("inf").
            init (float, optional): Initial amount. Defaults to 0.0.
            name (str, optional): The name of the store, if it doesn't exist as a state.
                Defaults to None.
        """
        super().__init__(env, capacity=capacity, init=init)
        self.name = name
        self._quantities: list[tuple[float, float]] = [(self._env.now, self._level)]

    def _record(self) -> None:
        reading = (self._env.now, self._level)
        if reading != self._quantities[-1]:
            self._quantities.append(reading)

    def _trigger_put(self, event: Event) -> None:  # type: ignore [override]
        super()._trigger_put(event)
        self._record()

    def _trigger_get(self, event: Event) -> None:  # type: ignore [override]
        super()._trigger_get(event)
        self._record()

    def _do_put(self, event: ContainerPut) -> bool | None:
        ans = super()._do_put(event)
        if event.triggered:
            self._record()
        return ans

    def _do_get(self, event: ContainerGet) -> bool | None:
        ans = super()._do_get(event)
        if event.triggered:
            self._record()
        return ans


class SelfMonitoringContinuousContainer(
    ContinuousContainer,
    MonitoringMixin,
    skip_classname=True,
):
    """A self-monitoring version of the Continuous Container."""

    def __init__(
        self,
        env: Environment,
        capacity: int | float,
        init: int | float = 0.0,
        error_empty: bool = True,
        error_full: bool = True,
        name: str | None = None,
    ) -> None:
        """A monitoring version of the Continuous container.

        The container allows continuous gets and puts.

        Args:
            env (Environment): SimPy Environment.
            capacity (int | float): Capacity of the container
            init (int | float, optional): Initial amount. Defaults to 0.0.
            error_empty (bool, optional): Error when it gets empty. Defaults to True.
            error_full (bool, optional): Error when it gets full. Defaults to True.
            name (str, optional): The name of the store, if it doesn't exist as a state.
                Defaults to None.
        """
        super().__init__(env, capacity, init, error_empty, error_full)
        self.name = name
        self._quantities = [(self._env.now, self._level)]

    def _set_level(self) -> float:
        """Set the level of the container based on the active gets/puts.

        Returns:
            float: The current level.
        """
        amt = super()._set_level()
        now = self._env.now
        if (now, amt) != self._quantities[-1]:
            self._quantities.append((now, amt))
        return amt


class SelfMonitoringSortedFilterStore(SortedFilterStore, SelfMonitoringStore, skip_classname=True):
    """A self-monitoring version of the SortedFilterStore."""

    def _do_get(self, event: _SortedFilterStoreGet) -> bool:  # type: ignore [override]
        ans = super()._do_get(event)
        if event.triggered:
            self._record("get")
        return ans


class SelfMonitoringReserveContainer(
    ReserveContainer,
    MonitoringMixin,
    skip_classname=True,
):
    """A self-monitoring version of the ReserveContainer."""

    def __init__(
        self,
        env: Environment,
        capacity: float = float("inf"),
        init: float = 0.0,
        name: str | None = None,
    ) -> None:
        """Create a store-like object that allows reservations, and records.

        Note that this store doesn't actually yield to SimPy when requesting.

        Use it to determine if anything is avaiable for reservation, but there is no
        queue for getting a reservation.

        Args:
            env (Environment): The SimPy Environment
            init (float, optional): Initial amount available. Defaults to 0.0.
            capacity (float, optional): Total capacity. Defaults to float("inf").
            name (str, optional): The name of the store, if it doesn't exist as a state.
                Defaults to None.
        """
        super().__init__(env, init, capacity)
        self.name = name
        self._quantities = [(env.now, init)]

    def _record(self) -> None:
        """Record the level of the store."""
        now = self._env.now
        data = (now, self._real_level)
        if data != self._quantities[-1]:
            self._quantities.append(data)

    def take(self, requester: Any) -> float:
        """Take some amount from the store, by a requester.

        Args:
            requester (Any): The entity requesting the amount

        Returns:
            float: The aount requested.
        """
        amt = super().take(requester)
        self._record()
        return amt

    def put(self, amount: float, capacity_increase: bool = False) -> None:
        """Put some amount into the store.

        Args:
            amount (float): The amount to put
            capacity_increase (bool, optional): Allow capacity to increase. Defaults to False.
        """
        super().put(amount, capacity_increase)
        self._record()



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from typing import Any

import upstage_des.api as UP
from upstage_des.type_help import TASK_GEN


class Mover(UP.Actor):
    fuel = UP.SharedLinearChangingState(recording=True)
    fuel_burn = UP.State[float](valid_types=(float,), frozen=True)
    location = UP.CartesianLocationChangingState(recording=True)
    speed = UP.State[float](valid_types=(float, int), recording=True)

    def get_distance(
        self, waypoints: list[UP.GeodeticLocation] | list[UP.CartesianLocation]
    ) -> float:
        d = waypoints[0] - self.location
        for i in range(1, len(waypoints)):
            d += waypoints[i] - waypoints[i - 1]
        return d


class Fly(UP.Task):
    def task(self, *, actor: Mover) -> TASK_GEN:
        destinations = list(self.get_actor_knowledge(actor, "destinations"))
        self.clear_actor_knowledge(actor, "destinations")
        actor.activate_state(
            state="location",
            task=self,
            speed=actor.speed,
            waypoints=destinations,
        )
        actor.activate_state(
            state="fuel",
            task=self,
            rate=actor.fuel_burn,
        )
        dist = actor.get_distance(destinations)
        time = dist / actor.speed
        yield UP.Wait(time)
        actor.deactivate_all_states(task=self)

    def on_interrupt(self, *, actor: Mover, cause: Any) -> UP.InterruptStates:
        # Allow subclassing to run a check prior to this
        # and input "restart" if they want.
        reason = None
        if isinstance(cause, UP.NucleusInterrupt):
            if cause.state_name == "speed":
                reason = "restart"
        if reason == "restart":
            rem_wypts = actor.get_remaining_waypoints(
                location_state="location",
            )
            self.set_actor_knowledge(
                actor,
                "destinations",
                rem_wypts,
                overwrite=True,
            )
            return self.INTERRUPT.RESTART
        return self.INTERRUPT.END


fly_factory = UP.TaskNetworkFactory.from_single_looping("Fly", Fly)
fly_end_factory = UP.TaskNetworkFactory.from_single_terminating("Fly", Fly)



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import pytest
import simpy as SIM
from simpy.resources import base
from simpy.resources.container import ContainerGet, ContainerPut
from simpy.resources.store import StoreGet, StorePut

from upstage_des.api import (
    Actor,
    EnvironmentContext,
    SimulationError,
    State,
    Task,
    add_stage_variable,
)
from upstage_des.events import (
    All,
    Any,
    BaseEvent,
    BaseRequestEvent,
    Event,
    Get,
    Put,
    ResourceHold,
    Wait,
)
from upstage_des.type_help import SIMPY_GEN, TASK_GEN


def test_base_event() -> None:
    init_time = 1.23
    with EnvironmentContext(initial_time=init_time) as env:
        base = BaseEvent()
        assert base.created_at == init_time, "Problem in environment time being stored in event"
        assert base.env is env, "Problem in environment being stored in event"

        with pytest.raises(NotImplementedError):
            base.as_event()


def test_wait_event() -> None:
    init_time = 1.23
    with EnvironmentContext(initial_time=init_time) as env:
        timeout = 1

        wait = Wait(timeout=timeout)
        assert wait.created_at == init_time, "Problem in environment time being stored in event"
        assert wait.env is env, "Problem in environment being stored in event"
        assert wait.timeout == timeout

        ret = wait.as_event()
        assert isinstance(ret, SIM.Timeout), "Wait doesn't return a simpy timeout"
        assert ret._delay == timeout, "Incorrect timeout time"

    with EnvironmentContext() as env:
        add_stage_variable("time_unit", "minutes")
        wait = Wait(timeout=1.1, timeout_unit="hours")
        assert wait.timeout == 1.1 * 60

    with EnvironmentContext(initial_time=init_time) as env:
        timeout_2 = [1, 3]
        wait = Wait.from_random_uniform(timeout_2[0], timeout_2[1])
        ret = wait.as_event()
        assert isinstance(ret, SIM.Timeout), "Wait doesn't return a simpy timeout"
        assert timeout_2[0] <= ret._delay <= timeout_2[1], "Incorrect timeout time"

        with pytest.raises(SimulationError):
            Wait(timeout={1, 2})  # type: ignore [arg-type]

        with pytest.raises(SimulationError):
            Wait(timeout="1")  # type: ignore [arg-type]

        with pytest.raises(SimulationError):
            Wait(timeout=[1])  # type: ignore [arg-type]

        with pytest.raises(SimulationError):
            Wait(timeout=[1, 2, 3])  # type: ignore [arg-type]


def test_base_request_event() -> None:
    init_time = 1.23
    with EnvironmentContext(initial_time=init_time) as env:
        base = BaseRequestEvent()
        assert base.created_at == init_time, "Problem in environment time being stored in event"
        assert base.env is env, "Problem in environment being stored in event"

        base.cancel()


def test_put_event_with_stores() -> None:
    with EnvironmentContext() as env:
        store = SIM.Store(env, capacity=1)
        put_object = ("A Test Object", 1.0)
        put_event = Put(store, put_object)

        assert put_event.calculate_time_to_complete() == 0.0, "Incorrect time to complete"
        returned_object = put_event.as_event()
        assert issubclass(returned_object.__class__, base.Put), (
            "Event returned is not simpy put event"
        )
        env.run()
        assert isinstance(returned_object, StorePut)
        assert returned_object.item is put_object, "Wrong object put"
        assert put_object in store.items

        put_object = ("A Second Test Object", 2.0)
        put_event = Put(store, put_object)
        event = put_event.as_event()
        env.run()
        assert not event.triggered, "Event shouldn't have completed"
        assert event in store.put_queue, "Event is not waiting"
        put_event.cancel()
        env.run()
        assert not event.triggered, "Event shouldn't have completed"
        assert event not in store.put_queue, "Event is still in the store's queue"


def test_put_event_with_containers() -> None:
    with EnvironmentContext() as env:
        container = SIM.Container(env, capacity=1)
        put_arg = 1.0
        put_event = Put(container, put_arg)

        assert put_event.calculate_time_to_complete() == 0.0, "Incorrect time to complete"
        returned_object = put_event.as_event()
        assert issubclass(returned_object.__class__, base.Put), (
            "Event returned is not simpy put event"
        )
        env.run()
        assert isinstance(returned_object, ContainerPut)
        assert returned_object.amount == put_arg, "Wrong amount put"
        assert container.level == put_arg

        put_arg = 2
        put_event = Put(container, put_arg)
        event = put_event.as_event()
        env.run()
        assert not event.triggered, "Event shouldn't have completed"
        assert event in container.put_queue, "Event is not waiting"
        put_event.cancel()
        env.run()
        assert not event.triggered, "Event shouldn't have completed"
        assert event not in container.put_queue, "Event is still in the store's queue"


def test_get_event_with_stores() -> None:
    with EnvironmentContext() as env:
        store = SIM.Store(env, capacity=1)
        put_object = ("A Test Object", 1.0)
        store.put(put_object)
        env.run()

        event = Get(store)
        assert event.calculate_time_to_complete() == 0.0, "Incorrect time to complete"
        returned_object = event.as_event()
        assert issubclass(returned_object.__class__, base.Get), (
            "Event returned is not simpy put event"
        )

        env.run()
        assert isinstance(event._request_event, StoreGet)
        item = event._request_event.value
        assert item is put_object, "Returned item is not the original item"
        item2 = event.get_value()
        assert item is item2, "Same object from both methods"

        event = Get(store)
        returned_object = event.as_event()
        env.run()
        assert returned_object in store.get_queue, "Event not in queue"
        event.cancel()
        assert returned_object not in store.get_queue, "Event is still in queue"


def test_get_event_with_containers() -> None:
    with EnvironmentContext() as env:
        container = SIM.Container(env, capacity=1)
        put_arg = 1.0
        container.put(put_arg)
        env.run()

        get_arg = 1.0
        event = Get(container, get_arg)
        assert event.calculate_time_to_complete() == 0.0, "Incorrect time to complete"
        returned_object = event.as_event()
        assert issubclass(returned_object.__class__, base.Get), (
            "Event returned is not simpy put event"
        )

        env.run()
        assert isinstance(event._request_event, ContainerGet)
        amount = event._request_event.amount
        assert amount == get_arg, "Returned item is not the original item"
        with pytest.raises(
            SimulationError,
            match="'get_value' is not supported for Containers. "
            "Check is_complete and use the amount you "
            "requested",
        ):
            event.get_value()

        event = Get(container, get_arg)
        returned_object = event.as_event()
        env.run()
        assert returned_object in container.get_queue, "Event not in queue"
        event.cancel()
        assert returned_object not in container.get_queue, "Event is still in queue"


def test_resource_events() -> None:
    with EnvironmentContext() as env:
        a_resource = SIM.Resource(env, capacity=1)

        request_object = ResourceHold(a_resource)
        assert request_object._stage == "request", "Request object in wrong state"
        request_object.as_event()
        env.run()

        assert request_object._stage == "release", "Request object in wrong state"
        assert a_resource.users[0] is request_object._request, "The user is the request object"

        new_request = ResourceHold(a_resource)
        assert new_request._stage == "request", "Request object in wrong state"
        new_request.as_event()
        env.run()

        assert new_request._stage == "release", "Request object in wrong state"
        assert new_request._request is not None
        assert not new_request._request.processed, "Request went through when it shouldn't"

        # put the old one back
        request_object.as_event()
        env.run()
        assert new_request._request is not None
        assert new_request._request.processed, "Follow-on request didn't go through"

        newest_request = ResourceHold(a_resource)
        assert newest_request._stage == "request", "Request object in wrong state"
        newest_request.as_event()
        env.run()

        # cancel it
        assert newest_request._stage == "release", "Request object in wrong state"
        assert newest_request._request is not None
        assert not newest_request._request.processed, "Request went through when it shouldn't"

        assert newest_request._request in a_resource.put_queue, (
            "Resource isn't waiting to be gathered"
        )
        with pytest.raises(SimulationError, match="Resource release requested.*?"):
            newest_request.as_event()

        newest_request.cancel()
        env.run()
        assert newest_request._request not in a_resource.put_queue, (
            "Resource hasn't left the wait queue"
        )


def test_multi_event() -> None:
    with EnvironmentContext() as env:
        event1 = Wait(1.0)
        event2 = Wait(1.5)
        event = All(event1, event2)
        assert event.calculate_time_to_complete() == 1.5

    with EnvironmentContext() as env:
        with pytest.warns(UserWarning):
            event1 = Wait(1.0)
            event3 = SIM.Timeout(env, 1.5)
            All(event1, event3)  # type: ignore [arg-type]


def test_and_event() -> None:
    with EnvironmentContext() as env:

        def run(env: SIM.Environment, data: dict[str, float]) -> SIMPY_GEN:
            event1 = Wait(1.0)
            event2 = Wait(1.5)

            event = All(event1, event2)
            yield event.as_event()
            data["time"] = env.now

        data: dict[str, float] = {}
        env.process(run(env, data))
        env.run()
        assert data["time"] == 1.5


def test_or_event() -> None:
    with EnvironmentContext() as env:
        data = {
            "time": 0.0,
        }

        def run(env: SIM.Environment) -> SIMPY_GEN:
            event1 = Wait(1.0)
            event2 = Wait(1.5)

            event = Any(event1, event2)
            yield event.as_event()
            data["time"] = env.now

        env.process(run(env))
        env.run()
        # SimPy still runs the simulation long enough to finish the timeout
        assert data["time"] == 1.0


def test_composite() -> None:
    with EnvironmentContext() as env:
        data = {
            "time": 0.0,
            "result": 0,
        }

        def run(env: SIM.Environment) -> SIMPY_GEN:
            event1 = Wait(1.0)
            event2 = Wait(1.5)

            event3 = Wait(2.1)
            event4 = Wait(0.9)

            event_a = Any(event1, event2)
            event_b = All(event3, event4, event_a)
            result = yield event_b.as_event()
            data["time"] = env.now
            data["result"] = len(result.events)

        env.process(run(env))
        env.run()
        assert data["time"] == 2.1
        assert data["result"] == 4


def test_process_in_multi() -> None:
    with EnvironmentContext() as env:

        def a_process() -> SIMPY_GEN:
            yield env.timeout(2)

        class Thing(Actor):
            result = State[dict]()
            events = State[list]()

        class TheTask(Task):
            def task(self, *, actor: Thing) -> TASK_GEN:
                wait = Wait(3.0)
                proc = env.process(a_process())
                res = yield Any(wait, proc)
                actor.events = [wait, proc]
                actor.result = res

        t = Thing(name="Thing", result=None, events=None)
        task = TheTask()
        task.run(actor=t)
        with pytest.warns(UserWarning):
            env.run()
        assert t.events[-1] in t.result
        assert t.events[0] not in t.result


def test_rehearse_process_in_multi() -> None:
    with EnvironmentContext() as env:

        def a_process() -> SIMPY_GEN:
            yield env.timeout(2)

        class Thing(Actor): ...

        class TheTask(Task):
            def task(self, *, actor: Thing) -> TASK_GEN:
                wait = Wait(3.0)
                proc = env.process(a_process())
                yield Any(wait, proc)

        t = Thing(name="Thing")
        task = TheTask()
        with pytest.raises(SimulationError, match="All events in a MultiEvent"):
            with pytest.warns(UserWarning):
                task.rehearse(actor=t)


def run_one(env: SIM.Environment, event: Event) -> SIMPY_GEN:
    yield env.timeout(1.0)
    event.succeed(data="here")


def run_two(env: SIM.Environment, event: Event, data: dict[str, float]) -> SIMPY_GEN:
    yield event._event
    data["time_two"] = env.now


def run_three(env: SIM.Environment, event: Event, data: dict[str, float]) -> SIMPY_GEN:
    yield event._event
    data["time_three"] = env.now


def run_four(env: SIM.Environment, event: Event) -> SIMPY_GEN:
    yield env.timeout(1.1)
    event.succeed()


def run_four_alt(env: SIM.Environment, event: Event) -> SIMPY_GEN:
    yield env.timeout(1.1)
    event.succeed()


def run_five(env: SIM.Environment, event: Event, data: dict[str, float]) -> SIMPY_GEN:
    # Timeout until after the event suceeded, but before its reset
    yield env.timeout(1.05)
    yield event.as_event()
    data["time_five"] = env.now


def test_basic_usage() -> None:
    with EnvironmentContext() as env:
        event = Event()
        assert event._event is not None
        assert isinstance(event._event, SIM.Event)
        assert event._event is event.as_event()
        assert event.is_complete() is False

        env.process(run_one(env, event))
        data: dict[str, float] = {}
        env.process(run_two(env, event, data))
        env.run()
        assert data["time_two"] == 1.0
        assert event.is_complete()
        payload = event.get_payload()
        assert payload == {"data": "here"}

        with pytest.raises(SimulationError):
            event.succeed()

        assert event.calculate_time_to_complete() == 0.0

        last_event = event._event
        event.reset()
        assert last_event is not event._event


def test_conflicts() -> None:
    data: dict[str, float] = {}
    with EnvironmentContext() as env:
        event = Event()
        env.process(run_one(env, event))
        env.process(run_two(env, event, data))
        env.process(run_three(env, event, data))
        env.run()
        assert data["time_two"] == data["time_three"]

    data: dict[str, float] = {}
    with EnvironmentContext() as env:
        with pytest.raises(SimulationError):
            event = Event()
            env.process(run_one(env, event))
            env.process(run_two(env, event, data))
            env.process(run_three(env, event, data))
            env.process(run_four(env, event))
            env.run()

    data: dict[str, float] = {}
    with EnvironmentContext() as env:
        event = Event()
        env.process(run_one(env, event))
        env.process(run_two(env, event, data))
        env.process(run_three(env, event, data))
        env.process(run_four_alt(env, event))
        env.process(run_five(env, event, data))
        env.run()
        assert data["time_two"] == data["time_three"]
        assert data["time_five"] == 1.1


def test_resubmit_wait_events() -> None:
    # Test the bug found in github issue 41
    # Where Wait wasn't acting like timeout for being resubmitted.

    # This illustrates the simpy behavior
    with EnvironmentContext() as env:
        w1 = Wait(1.1)
        w2 = Wait(2.2)

        # put the events into simpy
        w1.as_event()
        w2.as_event()

        env.run()
        assert env.now == 2.2

    # even when the timeout has no callbacks, it completes
    with EnvironmentContext() as env:
        w1 = Wait(1.1)
        w2 = Wait(2.2)

        def _proc() -> SIMPY_GEN:
            yield w1.as_event() | w2.as_event()
            assert env.now == 1.1

        env.process(_proc())
        env.run()
        assert env.now == 2.2

    # if we re-wait on w2, it should end at the right time.
    with EnvironmentContext() as env:
        w1 = Wait(1.1)
        w2 = Wait(2.2)

        def _proc() -> SIMPY_GEN:
            yield w1.as_event() | w2.as_event()
            assert env.now == 1.1
            yield w2.as_event()
            assert env.now == 2.2

        env.process(_proc())
        env.run()
        assert env.now == 2.2


def test_resubmit_get_put_events() -> None:
    # Make sure that get/put events don't hang.
    with EnvironmentContext() as env:
        store1 = SIM.Store(env)
        store2 = SIM.Store(env)

        def _put_stuff() -> SIMPY_GEN:
            yield Wait(1.0).as_event()
            yield store1.put("thing")
            yield Wait(1.0).as_event()
            yield store2.put("other")

        def _get_stuff() -> SIMPY_GEN:
            g1 = Get(store1)
            g2 = Get(store2)
            yield g1.as_event() | g2.as_event()
            assert g1.is_complete()
            assert g1.get_value() == "thing"
            assert env.now == 1.0
            yield g2.as_event()
            assert env.now == 2.0
            assert g2.is_complete()
            assert g2.get_value() == "other"

        env.process(_put_stuff())
        env.process(_get_stuff())
        env.run()

    # run the same through UPSTAGE
    with EnvironmentContext() as env:
        store1 = SIM.Store(env)
        store2 = SIM.Store(env)

        def _put_stuff() -> SIMPY_GEN:
            yield Wait(1.0).as_event()
            yield store1.put("thing")
            yield Wait(1.0).as_event()
            yield store2.put("other")

        def _get_stuff() -> SIMPY_GEN:
            g1 = Get(store1)
            g2 = Get(store2)
            yield Any(g1, g2).as_event()
            assert g1.is_complete()
            assert g1.get_value() == "thing"
            assert env.now == 1.0
            yield g2.as_event()
            assert env.now == 2.0
            assert g2.is_complete()
            assert g2.get_value() == "other"

        env.process(_put_stuff())
        env.process(_get_stuff())
        env.run()



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import pytest

from upstage_des.actor import Actor
from upstage_des.api import Task, UpstageError, Wait
from upstage_des.base import EnvironmentContext
from upstage_des.states import LinearChangingState, State
from upstage_des.type_help import TASK_GEN


class Piggy(Actor):
    a_level = LinearChangingState(recording=True)
    location = State[str]()


class Rider(Actor):
    b_level = LinearChangingState(recording=True)
    location = State[str]()
    piggy = State[Piggy]()


class RiderTask(Task):
    def task(self, *, actor: Rider) -> TASK_GEN:
        piggy = actor.piggy
        actor.activate_mimic_state(
            self_state="b_level",
            mimic_state="a_level",
            mimic_actor=piggy,
            task=self,
        )
        yield Wait(1.0)
        actor.deactivate_mimic_state(
            self_state="b_level",
            task=self,
        )


class RiderTaskTwo(Task):
    def task(self, *, actor: Rider) -> TASK_GEN:
        actor.activate_state(
            state="b_level",
            rate=2,
            task=self,
        )
        yield Wait(5.0)
        actor.deactivate_all_states(task=self)


def test_simple() -> None:
    with EnvironmentContext() as env:
        pig = Piggy(name="Piggy", a_level=1, location="here")
        ride = Rider(name="Rider", b_level=2, location="there", piggy=pig)

        assert ride.location == "there"
        ride.activate_mimic_state(
            self_state="location",
            mimic_state="location",
            mimic_actor=pig,
            task="None",  # type: ignore [arg-type]
        )
        assert ride.location == "here"
        pig.location = "way over there"
        assert ride.location == "way over there"

        ride.deactivate_mimic_state(self_state="location", task="None")  # type: ignore [arg-type]
        assert ride.location == "way over there"

        # a later test will check this differently
        assert not hasattr(ride, "_location_history")

        assert ride.b_level == 2
        ride.activate_mimic_state(
            self_state="b_level",
            mimic_state="a_level",
            mimic_actor=pig,
            task="None",  # type: ignore [arg-type]
        )
        pig.activate_state(
            state="a_level",
            rate=2,
            task=None,  # type: ignore [arg-type]
        )
        assert ride.b_level == 1
        env.run(until=4)

        pig.deactivate_all_states(task=None)  # type: ignore [arg-type]

        assert ride.b_level == 9
        ride.deactivate_all_mimic_states(task="None")  # type: ignore [arg-type]
        assert ride.b_level == 9

        ride.activate_state(
            state="b_level",
            rate=1,
            task=None,  # type: ignore [arg-type]
        )
        env.run(until=5)
        assert ride.b_level == 10
        assert pig.a_level == 9


def test_rehearsing() -> None:
    with EnvironmentContext():
        pig = Piggy(name="Piggy", a_level=1, location="here")
        ride = Rider(name="Rider", b_level=2, location="there", piggy=pig)

        task = RiderTask()
        with pytest.raises(UpstageError):
            task.rehearse(actor=ride)


def test_rehearse_from_piggyback() -> None:
    with EnvironmentContext() as env:
        pig = Piggy(name="Piggy", a_level=1, location="here")
        ride = Rider(name="Rider", b_level=2, location="there", piggy=pig)

        assert ride.b_level == 2
        ride.activate_mimic_state(
            self_state="b_level",
            mimic_state="a_level",
            mimic_actor=pig,
            task="None",  # type: ignore [arg-type]
        )
        pig.activate_state(
            state="a_level",
            rate=2,
            task=None,  # type: ignore [arg-type]
        )
        assert ride.b_level == 1
        env.run(until=4)
        pig.deactivate_all_states(task=None)  # type: ignore [arg-type]

        # Do not deactivate, instead call RiderTaskTwo
        # Pig went a_level to 9
        task = RiderTaskTwo()
        new_ride = task.rehearse(actor=ride)
        assert new_ride.b_level == 19
        assert ride.b_level == 9
        assert pig.a_level == 9

        ride.deactivate_mimic_state(self_state="b_level", task="None")  # type: ignore [arg-type]
        pig.a_level = 12
        assert new_ride.b_level == 19
        assert ride.b_level == 9


def test_double_mimic() -> None:
    with EnvironmentContext():
        pig = Piggy(name="Piggy", a_level=1, location="here")
        ride = Rider(name="Rider", b_level=2, location="there", piggy=pig)

        ride.activate_mimic_state(
            self_state="b_level",
            mimic_state="a_level",
            mimic_actor=pig,
            task="None",  # type: ignore [arg-type]
        )

        with pytest.raises(UpstageError):
            ride.activate_mimic_state(
                self_state="b_level",
                mimic_state="a_level",
                mimic_actor=pig,
                task="None",  # type: ignore [arg-type]
            )


def test_interrupt_deactivate() -> None:
    with EnvironmentContext() as env:
        pig = Piggy(name="Piggy", a_level=1, location="here")
        ride = Rider(name="Rider", b_level=2, location="there", piggy=pig)

        ride.activate_mimic_state(
            self_state="b_level",
            mimic_state="a_level",
            mimic_actor=pig,
            task="None",  # type: ignore [arg-type]
        )
        task = RiderTaskTwo()
        task.run(actor=ride)
        env.run()


def test_record() -> None:
    with EnvironmentContext():
        pig = Piggy(name="Piggy", a_level=1, location="here")

        class Rec(Actor):
            state = State[int](recording=True)

        class Rec2(Actor):
            state = State[int](recording=False)

        ride = Rec(
            name="another rider",
            state=3,
        )
        ride1 = Rec(
            name="another rider a",
            state=3,
        )
        ride2 = Rec2(
            name="another rider 2",
            state=3,
        )

        ride.activate_mimic_state(
            self_state="state",
            mimic_state="a_level",
            mimic_actor=pig,
            task="None",  # type: ignore [arg-type]
        )

        ride1.activate_mimic_state(
            self_state="state",
            mimic_state="a_level",
            mimic_actor=pig,
            task="None",  # type: ignore [arg-type]
        )

        ride2.activate_mimic_state(
            self_state="state",
            mimic_state="a_level",
            mimic_actor=pig,
            task="None",  # type: ignore [arg-type]
        )

        pig.a_level = 23
        assert ride.state == 23
        assert ride1.state == 23
        assert ride2.state == 23
        assert "state" in ride._state_histories
        assert "state" in ride1._state_histories
        assert "state" not in ride2._state_histories
        assert ride._state_histories["state"][1] == (0, 23)
        assert ride1._state_histories["state"][1] == (0, 23)



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from inspect import isgeneratorfunction
from typing import Any, cast

import pytest
from simpy import Environment, Process

from upstage_des.actor import Actor
from upstage_des.api import InterruptStates, SimulationError
from upstage_des.base import EnvironmentContext
from upstage_des.events import Wait
from upstage_des.states import LinearChangingState, State
from upstage_des.task import Task, TerminalTask
from upstage_des.type_help import SIMPY_GEN, TASK_GEN


class ActorForTest(Actor):
    dummy = State[float](recording=True)


class ActorChangeForTest(Actor):
    dummy = LinearChangingState()


class WorkingTask(Task):
    times: list[float]

    def task(self, *, actor: ActorForTest) -> TASK_GEN:
        for wait_period in self.times:
            the_event = Wait(wait_period)
            yield the_event
            actor.dummy += wait_period


class ChangingTask(Task):
    times: list[float]
    rate: float

    def task(self, *, actor: ActorForTest) -> TASK_GEN:
        for t in self.times:
            the_event = Wait(t)
            actor.activate_state(state="dummy", task=self, rate=self.rate)
            yield the_event
            actor.deactivate_state(state="dummy", task=self)


class Actor2Test(Actor):
    dummy = State[Any]()


class WorkingTask2(Task):
    times: list[float]
    log: list[str]

    def task(self, *, actor: ActorChangeForTest) -> TASK_GEN:
        for wait_period in self.times:
            wait_event = Wait(wait_period)
            self.log.append(
                f"{self.env.now}: {self.__class__.__name__} "
                f"waiting {wait_period}, value={actor.dummy}"
            )
            yield wait_event
            self.log.append(
                f"{self.env.now}: {self.__class__.__name__} "
                f"finished waiting {wait_period}, "
                f"value={actor.dummy}"
            )
            actor.dummy += wait_period


class ChangingTask2(Task):
    times: list[float]
    rate: float
    log: list[str]

    def task(self, *, actor: ActorForTest | ActorChangeForTest) -> TASK_GEN:
        for wait_period in self.times:
            wait_event = Wait(wait_period)
            actor.activate_state(state="dummy", task=self, rate=self.rate)
            actor.set_knowledge("example for logging", "a value", overwrite=True)
            self.log.append(
                f"{self.env.now}: {self.__class__.__name__} "
                f"waiting {wait_period}, value={actor.dummy}"
            )
            yield wait_event
            self.log.append(
                f"{self.env.now}: {self.__class__.__name__} finished "
                f"waiting {wait_period}, value={actor.dummy}"
            )
            actor.deactivate_state(state="dummy", task=self)


def _task_runner(env: Environment, rate: float, timeout_point: float) -> SIMPY_GEN:
    use_actor = ActorChangeForTest(name="testing", dummy=0.0, debug_log=True)
    times = [1.0, 2.0]

    task_object = ChangingTask2()
    task_object.times = times
    task_object.rate = rate
    task_object.log = []

    task_generator = task_object.run(
        actor=use_actor,
    )
    timeout = env.timeout(timeout_point)

    yield task_generator | timeout

    if task_generator.is_alive:
        task_generator.interrupt("cancelling")

    return use_actor


def test_creation() -> None:
    with EnvironmentContext():
        _ = WorkingTask()


def test_failures_for_tasks_with_simpy_events() -> None:
    with EnvironmentContext() as env:
        actor = ActorForTest(name="testing", dummy=0)

        class BrokenTask(Task):
            def task(self, *, actor: ActorForTest) -> TASK_GEN:
                yield self.env.timeout(1.0)  # type: ignore [misc, union-attr]

        # msg = "*Task is yielding objects without `as_event`*"
        with pytest.raises(SimulationError):  # , match=msg):
            the_task = BrokenTask()
            _ = the_task.run(
                actor=actor,
            )
            env.run()

        # msg = "*'MockEnvironment' object has no attribute 'timeout'*"
        with pytest.raises(AttributeError):  # , match=msg):
            the_task = BrokenTask()
            the_task.rehearse(
                actor=actor,
            )


def test_failures_for_tasks_with_incorrect_events() -> None:
    with EnvironmentContext():
        actor = ActorForTest(name="testing", dummy=0)

        class BlankEvent:
            def __init__(self, **kwargs: Any) -> None:
                pass

        class BrokenTask(Task):
            event_class: type

            def task(self, *, actor: ActorForTest) -> TASK_GEN:
                yield self.event_class()

        # msg = '*must be a subclass of BaseEvent*'
        with pytest.raises(SimulationError):
            task_instance = BrokenTask()
            task_instance.event_class = BlankEvent
            task_instance.rehearse(
                actor=actor,
            )


def test_running_as_rehearsal() -> None:
    with EnvironmentContext() as env:
        actor = ActorForTest(name="testing", dummy=0)
        times = [1.0, 2.0]
        task_object = WorkingTask()
        task_object.times = times
        rehearse_function = task_object.rehearse

        # assert that the task is not a generator
        msg = "Task when tested should not be a generator"
        assert not isgeneratorfunction(rehearse_function), msg

        result = rehearse_function(
            actor=actor,
        )
        msg = "Result of testing task must be an actor"
        assert isinstance(result, Actor), msg

        assert env.now == 0, "Environment time must not increase"

        msg = "Actor returned by task test is missing expected state changes"
        assert result.dummy == 3, msg

        msg = "Actor returned by task test needs to keep recorded data"
        assert len(result._state_histories["dummy"]) == 3, msg
        assert result._state_histories["dummy"][0] == (0, 0), msg
        assert result._state_histories["dummy"][2] == (sum(times), sum(times)), msg

        msg = "No linkage between original actor and dummy actor"
        assert actor.dummy == 0, msg

        msg = "Environment not properly reset"
        assert task_object.env is env, msg


def test_running() -> None:
    with EnvironmentContext() as env:
        actor = ActorForTest(name="testing", dummy=0)
        times = [1.0, 2.0]

        task_object = WorkingTask()
        task_object.times = times
        task_process = task_object.run(
            actor=actor,
        )
        env.run()
        assert env.now == 3, "Environment time must increase"
        assert actor.dummy == 3, "Actor state must change"
        assert isinstance(task_process, Process), "Task process is not an instance of simpy.Process"


def test_interrupting() -> None:
    with EnvironmentContext() as env:
        timeout_point = 0.5
        rate = 0.5
        proc = env.process(
            _task_runner(
                env,
                rate=rate,
                timeout_point=timeout_point,
            )
        )
        env.run()
        actor = cast(ActorForTest, proc.value)
        msg = "Task interruption ended at the wrong time"
        assert actor.dummy == timeout_point * rate, msg


def test_interrupting_two() -> None:
    # Do the timeout right when a time will end
    with EnvironmentContext() as env:
        timeout_point = 1.0
        rate = 3.5
        proc = env.process(
            _task_runner(
                env=env,
                rate=rate,
                timeout_point=timeout_point,
            )
        )
        env.run()
        actor = cast(ActorForTest, proc.value)
        msg = "Task interruption ended at the wrong time"
        assert actor.dummy == timeout_point * rate, msg


def test_simultaneous_task() -> None:
    with EnvironmentContext() as env:
        actor = ActorChangeForTest(name="testing", dummy=0.0)

        def task_runner(
            *,
            task_class: type[WorkingTask2 | ChangingTask2],
            interrupt_time: float,
            **task_kwargs: Any,
        ) -> SIMPY_GEN:
            task = task_class()
            task.log = []
            for k, v in task_kwargs.items():
                setattr(task, k, v)
            running_task = task.run(
                actor=actor,
            )

            timeout = env.timeout(interrupt_time)

            yield running_task | timeout

            if running_task.is_alive:
                running_task.interrupt("cancelling")

            return actor

        _ = env.process(
            task_runner(
                task_class=ChangingTask2,
                rate=1.0,
                times=[1.0, 3.0, 5.0, 6.0],
                interrupt_time=10.0,
            )
        )

        _ = env.process(
            task_runner(
                task_class=WorkingTask2,
                times=[2.0, 2.0, 2.0, 10.0],
                interrupt_time=12.0,
            )
        )

        env.run(until=20.0)


def test_terminal_task_run(
    task_objects: tuple[type[TerminalTask], type[TerminalTask], type[Actor]],
) -> None:
    EndPoint, EndPointBase, Dummy = task_objects

    with EnvironmentContext() as env:
        actor = Dummy(name="x", status="Good", debug_log=True)
        task = EndPoint()

        proc = task.run(actor=actor)
        env.run()
        assert env.now == 0

        assert "The Message" in actor._debug_log[-1][1]

        with pytest.raises(SimulationError, match=".+Cannot interrupt a terminal.+"):
            proc.interrupt()
            env.run()

        actor = Dummy(name="x", status="Good", debug_log=True)
        task = EndPointBase()
        proc = task.run(actor=actor)
        env.run()
        assert env.now == 0

        assert "Entering terminal task:" in actor._debug_log[-1][1]


def test_terminal_task_rehearse(
    task_objects: tuple[type[TerminalTask], type[TerminalTask], type[Actor]],
) -> None:
    EndPoint, _, Dummy = task_objects
    with EnvironmentContext():
        actor = Dummy(name="x", status="Good")
        task = EndPoint()

        clone = task.rehearse(actor=actor)
        assert clone.env.now == task._time_to_complete


class Dummy(Actor):
    status = State[str]()
    rate = State[float]()
    changer = LinearChangingState(recording=True)


class Restartable(Task):
    def task(self, *, actor: Dummy) -> TASK_GEN:
        actor.activate_state(
            state="changer",
            task=self,
            rate=actor.rate,
        )
        self.set_marker("change to test")
        yield Wait(10.0)
        actor.deactivate_all_states(task=self)

    def on_interrupt(self, *, actor: Dummy, cause: Any) -> InterruptStates:
        if cause == "restart":
            return self.INTERRUPT.RESTART
        else:
            return self.INTERRUPT.END


def test_restart() -> None:
    with EnvironmentContext() as env:
        actor = Dummy(
            name="Example",
            status="available",
            rate=2.3,
            changer=0.0,
            debug_log=True,
        )

        task = Restartable()
        proc = task.run(actor=actor)
        env.run(until=3.4)
        proc.interrupt(cause="restart")
        env.run()
        assert pytest.approx(actor.changer) == 2.3 * (3.4 + 10)



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import simpy as SIM

import upstage_des.api as UP
from upstage_des.type_help import TASK_GEN

from .mothership import Mothership
from .mover import Mover


class Flyer(Mover):
    fuel_capacity = UP.State[float](valid_types=(float,), frozen=True)
    fuel_draw = UP.State[float](valid_types=(int, float), frozen=True)
    messages = UP.ResourceState[SIM.Store](default=SIM.Store, valid_types=(SIM.Store,))
    approach = UP.State[bool](default=False, valid_types=(bool,), recording=True)


class MissionPlanning(UP.Task):
    def task(self, *, actor: Flyer) -> TASK_GEN:
        # Figure out the time to reach the right waypoint
        # Schedule the approach network
        # and a speed change for that time.
        # Also, kick off the flying.
        refuel_point = self.get_actor_knowledge(actor, "meetup", must_exist=True)
        time_to_point = (actor.location - refuel_point) / actor.speed
        yield UP.Wait(time_to_point)
        actor.approach = True


mission_plan_net = UP.TaskNetworkFactory.from_single_terminating("plan", MissionPlanning)


class ApproachWait(UP.Task):
    def task(self, *, actor: Flyer) -> TASK_GEN:
        yield UP.Event()

    def on_interrupt(self, *, actor: Flyer, cause: UP.NucleusInterrupt) -> UP.InterruptStates:
        assert cause.state_name == "approach"
        return self.INTERRUPT.END


class ApproachMothership(UP.Task):
    def task(self, *, actor: Flyer) -> TASK_GEN:
        the_mothership = self.get_actor_knowledge(actor, "mothership", must_exist=True)
        refuel_speed = self.get_actor_knowledge(actor, "refuel_speed", must_exist=True)
        actor.speed = refuel_speed  # NUCLEUS INTERACTION
        yield UP.Put(the_mothership.messages, (actor, -actor.fuel_draw))
        proceed = yield UP.Get(actor.messages)
        assert proceed == "GO"


class Refuel(UP.Task):
    def task(self, *, actor: Flyer) -> TASK_GEN:
        # ignore any particular timing to "full fuel"
        needed = actor.fuel_capacity - actor.fuel
        # we're here if we got the message to go
        actor.activate_state(
            state="fuel",
            task=self,
            rate=actor.fuel_draw,
        )
        time = needed / actor.fuel_draw
        yield UP.Wait(time)
        actor.deactivate_state(state="fuel", task=self)
        the_mothership: Mothership = self.get_actor_knowledge(actor, "mothership", must_exist=True)
        yield UP.Put(the_mothership.messages, (actor, 0))
        self.set_actor_knowledge(actor, "done_refueling", True, overwrite=True)


flyer_refuel_factory = UP.TaskNetworkFactory.from_ordered_loop(
    "flyer_refuel",
    [ApproachWait, ApproachMothership, Refuel],
)



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import pytest

from upstage_des.api import (
    EnvironmentContext,
    UpstageBase,
    UpstageError,
    add_stage_variable,
    get_stage,
    get_stage_variable,
)
from upstage_des.base import clear_top_context, create_top_context


class Stager(UpstageBase): ...


def test_stage() -> None:
    with EnvironmentContext():
        source = Stager()
        # Complain when accessing an unset attribute
        with pytest.raises(AttributeError):
            source.stage.stage_model

        # setting with the method
        add_stage_variable("stage_model", 1)
        assert source.stage.stage_model == 1

        # Setting without the method
        add_stage_variable("altitude_units", 2)
        assert source.stage.altitude_units == 2

        # Setting should yell after a set
        with pytest.raises(UpstageError):
            add_stage_variable("altitude_units", 3)

    # After the context, it should not exists
    with pytest.raises(UpstageError):
        source.stage


def test_contextless_stage() -> None:
    ctx = create_top_context()
    add_stage_variable("example", 1.234)

    assert get_stage_variable("example") == 1.234

    stage = get_stage()
    assert stage.example == 1.234

    # dropping into a new context ignores the above
    with EnvironmentContext():
        add_stage_variable("example", 8.675)
        assert get_stage_variable("example") == 8.675

    assert get_stage_variable("example") == 1.234

    clear_top_context(ctx)

    with pytest.raises(ValueError, match="Stage should have been set."):
        get_stage_variable("example")


if __name__ == "__main__":
    test_contextless_stage()



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""Tasks constitute the actions that Actors can perform."""

from collections.abc import Callable, Generator, Iterable
from enum import IntFlag
from functools import wraps
from typing import TYPE_CHECKING, Any, TypeVar
from warnings import warn

from simpy import Environment as SimpyEnv
from simpy import Event as SimpyEvent
from simpy import Interrupt, Process

if TYPE_CHECKING:
    from .actor import Actor
    from .task_network import TaskNetwork

from .base import ENV_CONTEXT_VAR, MockEnvironment, SettableEnv, SimulationError
from .constants import PLANNING_FACTOR_OBJECT
from .events import BaseEvent, Event

TASK_TYPE = Generator[BaseEvent | Process, Any, None]


__all__ = ("DecisionTask", "Task", "process", "TerminalTask", "TASK_TYPE", "InterruptStates")


NOT_IMPLEMENTED_MSG = "User must define the actions performed when executing this task"


class InterruptStates(IntFlag):
    """Class that describes how to behave after an interrupt."""

    END = 0
    IGNORE = 1
    RESTART = 2


def process(
    func: Callable[..., Generator[SimpyEvent, Any, None]],
) -> Callable[..., Process]:
    """Decorate a ``simpy`` process to schedule it as a callable.

    Allows users to decorate a generator, and when they want to schedule them
    as a ``simpy`` process, they can simply call it, e.g., instead of calling:

    Usage:

    >>> from upstage.api import process, Wait
    ...
    >>> @process
    >>> def generator(wait_period=1.0, msg="Finished Waiting"):
    >>>     # A simple process that periodically prints a statement
    >>>     while True:
    >>>         yield Wait(wait_period).as_event()
    >>>         print(msg)
    ...
    >>> @process
    >>> def another_process():
    >>>     # Some other process that calls the first one
    >>>     generator()

    Args:
        func (Callable[..., Generator[BaseEvent, None, None]]): The process function that is a
        generator of simpy events.

    Returns:
        Process: The generator as a ``simpy`` process.

    Note:
        The value of this decorator is that it reduces the chance of a user
        forgetting to call the generator as a process, which tends to produce
        behaviors that are difficult to troubleshoot because the code will
        build and can run, but the simulation will not work schedule the
        process defined by the generator.

    """

    @wraps(func)
    def wrapped_generator(*args: Any, **kwargs: Any) -> Process:
        """Wrap the generator with a function that calls it as a process."""
        try:
            environment = ENV_CONTEXT_VAR.get()
        except LookupError:
            raise SimulationError("No environment found on process call")
        return environment.process(func(*args, **kwargs))

    return wrapped_generator


EVT = TypeVar("EVT", bound=BaseEvent)
REH_ACTOR = TypeVar("REH_ACTOR", bound="Actor")


class Task(SettableEnv):
    """A Task is an action that can be performed by an Actor."""

    INTERRUPT = InterruptStates

    def __init__(self) -> None:
        """Create a task instance."""
        super().__init__()
        self._proc: TASK_TYPE | None = None
        self._network_name: str | None = None
        self._network_ref: TaskNetwork | None = None
        self._marker: str | None = None
        self._marked_time: float | None = None
        self._interrupt_action: InterruptStates = InterruptStates.END
        self._rehearsing: bool = False

    def task(self, *, actor: Any) -> TASK_TYPE:
        """Define the process this task follows."""
        raise NotImplementedError(NOT_IMPLEMENTED_MSG)

    def on_interrupt(self, *, actor: Any, cause: Any) -> InterruptStates:
        """Define any actions to take on the actor if this task is interrupted.

        Note:
            Custom Tasks can overwrite this method so they can handle being
            interrupted with a custom procedure. By default, interrupt ends the
            task.

        Args:
            actor (Actor): the actor using the task
            cause (Any): Optional data for the interrupt
        """
        actor.log(f"Interrupted while performing {self}. Reasons: {cause}")
        return self._interrupt_action

    def set_marker(
        self, marker: str, interrupt_action: InterruptStates = InterruptStates.END
    ) -> None:
        """Set a marker to help with inspection of interrupts.

        The interrupt_action is set for when no `on_interrupt` is implemented.

        Args:
            marker (str): String for the marker.
            interrupt_action (InterruptStates, optional): Action to take on interrupt.
            Defaults to InterruptStates.END.
        """
        self._marker = marker
        self._marked_time = self.env.now
        self._interrupt_action = interrupt_action

    def get_marker(self) -> str | None:
        """Get the current marker.

        Returns:
            str | None: Marker (or None if cleared)
        """
        return self._marker

    def get_marker_time(self) -> float | None:
        """The time the current marker was set.

        Returns:
            float | None: Marker set time (or None if cleared)
        """
        return self._marked_time

    def clear_marker(self) -> None:
        """Clear the marker and set that an interrupt ends the task."""
        self._marker = None
        self._marked_time = None
        self._interrupt_action = InterruptStates.END

    def _set_network_ref(self, network: "TaskNetwork") -> None:
        """Set the reference to the task network object.

        Args:
            network (TaskNetwork): The network
        """
        if self._network_ref is not None:
            raise SimulationError(
                "Setting task network reference on task that already has a network"
            )
        self._network_ref = network

    def _set_network_name(self, network_name: str) -> None:
        """Set the name of the network this task is in.

        Args:
            network_name (str): Network name
        """
        if self._network_name is not None:
            raise SimulationError("Setting task network name on task that already has a network")
        self._network_name = network_name

    def clear_actor_task_queue(self, actor: "Actor") -> None:
        """Clear out the task queue on the network.

        Args:
            actor (Actor): The actor whose queue will be cleared
        """
        assert self._network_name is not None
        actor.clear_task_queue(self._network_name)

    def set_actor_task_queue(self, actor: "Actor", task_list: list[str]) -> None:
        """Set the task queue on the actor.

        This assumes an empty queue.

        Args:
            actor (Actor): The actor to modify the task queue of
            task_list (list[str]): The list of task names to queue.
        """
        assert self._network_name is not None
        actor.set_task_queue(self._network_name, task_list)

    def get_actor_task_queue(self, actor: "Actor") -> list[str]:
        """Get the task queue on the actor.

        Args:
            actor (Actor): The actor to modify the task queue of
        """
        assert self._network_name is not None
        return actor.get_task_queue(self._network_name)

    def get_actor_next_task(self, actor: "Actor") -> str | None:
        """Get the next queued task.

        Args:
            actor (Actor): The actor to get the next task from

        Returns:
            str | None: The next task name (or None if no task)
        """
        assert self._network_name is not None
        return actor.get_next_task(self._network_name)

    def set_actor_knowledge(
        self,
        actor: "Actor",
        name: str,
        value: Any,
        overwrite: bool = False,
    ) -> None:
        """Set knowledge on the actor.

        Convenience method for passing in the name of task for actor logging.

        Args:
            actor (Actor): The actor to set knowledge on.
            name (str): Name of the knowledge
            value (Any): Value of the knowledge
            overwrite (bool, optional): Allow overwrite or not. Defaults to False.
        """
        cname = self.__class__.__qualname__
        actor.set_knowledge(name, value, overwrite=overwrite, caller=cname)

    def clear_actor_knowledge(self, actor: "Actor", name: str) -> None:
        """Clear knowledge from an actor.

        Convenience method for passing in the name of task for actor logging.

        Args:
            actor (Actor): The actor to clear knowledge from
            name (str): The name of the knowledge
        """
        cname = self.__class__.__qualname__
        actor.clear_knowledge(name, caller=cname)

    @staticmethod
    def get_actor_knowledge(actor: "Actor", name: str, must_exist: bool = False) -> Any:
        """Get knowledge from the actor.

        Args:
            actor (Actor): The actor to get knowledge from.
            name (str): Name of the knowledge
            must_exist (bool, optional): Raise errors if the knowledge doesn't exist.
            Defaults to False.

        Returns:
            Any: The knowledge value, which could be None
        """
        return actor.get_knowledge(name, must_exist)

    def get_and_clear_actor_knowledge(self, actor: "Actor", name: str) -> Any:
        """Get and clear knowledge on an actor.

        The knowledge is assumed to exist.

        Args:
            actor (Actor): The actor to get knowledge from.
            name (str): The knowledge name.

        Returns:
            Any: The knowledge value.
        """
        cname = self.__class__.__qualname__
        return actor.get_and_clear_knowledge(name, caller=cname)

    def set_actor_bulk_knowledge(
        self, actor: "Actor", know: dict[str, Any], overwrite: bool = False
    ) -> None:
        """Set multiple knowledge entries at once.

        Args:
            actor (Actor): The actor to operate on.
            know (dict[str, Any]): Dictionary of key:value pairs of knowledge.
            overwrite (bool, optional): If overwrite is allowed. Defaults to False.
        """
        for k, v in know.items():
            self.set_actor_knowledge(actor, k, v, overwrite)

    def clear_actor_bulk_knowledge(self, actor: "Actor", names: Iterable[str]) -> None:
        """Clear a list of knowledge entries.

        Args:
            actor (Actor): The actor to operate on.
            names (Iterable[str]): Knowledge names.
        """
        for name in names:
            self.clear_actor_knowledge(actor, name)

    def get_actor_bulk_knowledge(
        self, actor: "Actor", names: Iterable[str], must_exist: bool = False
    ) -> dict[str, Any]:
        """Get multiple knowledge items.

        Args:
            actor (Actor): The actor to operate on.
            names (Iterable[str]): Names of the knowledge
            must_exist (bool, optional): If all entires must exist. Defaults to False.

        Returns:
            dict[str, Any]: The knowledge values. None if not present.
        """
        return {name: self.get_actor_knowledge(actor, name, must_exist) for name in names}

    def get_and_clear_actor_bulk_knowledge(
        self, actor: "Actor", names: Iterable[str], caller: str | None = None
    ) -> dict[str, Any]:
        """Get and clear multiple knowledge entries.

        Args:
            actor (Actor): The actor to operate on.
            names (Iterable[str]): The knowledge to retrieve and delete.
            caller (str | None, optional): The name of the caller. Defaults to None.

        Returns:
            dict[str, Any]: The retrieved knowledge.
        """
        return {name: self.get_and_clear_actor_knowledge(actor, name) for name in names}

    def _clone_actor(self, actor: REH_ACTOR, knowledge: dict[str, Any] | None) -> REH_ACTOR:
        """Create a clone of the actor.

        Args:
            actor (Actor): The actor to clone
            knowledge (dict[str, Any] | None): Additional knowledge to add.

        Returns:
            Actor: Cloned actor
        """
        mocked_env = MockEnvironment.mock(self.env)
        self.env = mocked_env
        understudy = actor.clone(
            new_env=mocked_env,
            knowledge=knowledge,
        )
        return understudy

    def rehearse(
        self,
        *,
        actor: REH_ACTOR,
        knowledge: dict[str, Any] | None = None,
        cloned_actor: bool = False,
        **kwargs: Any,
    ) -> REH_ACTOR:
        """Rehearse the task to evaluate its feasibility.

        Args:
            actor (Actor): The actor to rehearse in the task
            knowledge (dict[str, Any], optional): Knowledge to add to the actor. Defaults to None.
            cloned_actor (bool, optional): If the actor is a clone or not. Defaults to False.
            kwargs (Any): Optional args to send to the task.

        Returns:
            Actor: The cloned actor with a state reflecting the task flow.
        """
        knowledge = {} if knowledge is None else knowledge
        _old_env = self.env
        understudy = actor
        if not cloned_actor:
            understudy = self._clone_actor(actor, knowledge)
        if not isinstance(understudy.env, MockEnvironment):
            raise SimulationError("Bad actor cloning.")
        self.env = understudy.env
        mocked_env: MockEnvironment = understudy.env

        self._rehearsing = True
        generator = self.task(actor=understudy, **kwargs)
        returned_item = None
        while True:
            try:
                if returned_item is None:
                    next_event = next(generator)
                else:
                    next_event = generator.send(returned_item)
                    returned_item = None
                if not issubclass(next_event.__class__, BaseEvent):
                    msg = f"Task {self} event {next_event}"
                    if isinstance(next_event, Process):
                        raise SimulationError(msg + " cannot be a process during rehearsal.")
                    raise SimulationError(msg + " must be a subclass of BaseEvent!")
                time_advance, returned_item = next_event.rehearse()
                mocked_env.now += time_advance

            except StopIteration:
                # warn(f"Stopping rehearsal of task '{self.__class__.__name__}' "
                #      f"for actor '{actor}'! [Rehearsal duration: "
                #      f"{self.env.now - _old_env.now:.3g}]")
                break

        self.env = _old_env
        self._rehearsing = False
        return understudy

    def _handle_interruption(
        self, actor: "Actor", interrupt: Interrupt, next_event: BaseEvent | Process
    ) -> tuple[bool, bool]:
        """Clean up after an interrupt and perform interrupt checks/actions.

        Args:
            actor (Actor): _description_
            interrupt (Interrupt): _description_
            next_event (BaseEvent): _description_

        Returns:
            bool: If the task should be stopped
            bool: If the task should be restarted
        """
        # test the interrupt behavior:
        stop_run = False
        restart = False
        _interrupt_action = self.on_interrupt(
            actor=actor,
            cause=interrupt.cause,
        )
        if _interrupt_action is None:
            raise SimulationError("No interrupt behavior returned from `on_interrupt`")

        if _interrupt_action in (InterruptStates.END, InterruptStates.RESTART):
            actor.log(f"Interrupted by {interrupt}.")
            actor.deactivate_all_states(task=self)
            actor.deactivate_all_mimic_states(task=self)
            if isinstance(next_event, BaseEvent):
                names = list(actor._knowledge.keys())
                for name in names:
                    if actor._knowledge[name] is next_event:
                        actor.clear_knowledge(name, caller=f"Clearing knowledge event {name}")
                next_event.cancel()
            elif isinstance(next_event, Process):
                next_event.interrupt(cause=interrupt.cause)
            else:
                raise SimulationError(f"Bad event passed: {next_event}")
            stop_run = True
            if _interrupt_action is InterruptStates.RESTART:
                restart = True
        elif _interrupt_action is InterruptStates.IGNORE:
            # go back to waiting on the event
            stop_run = False
        else:
            raise SimulationError(f"Wrong interrupt action value: {_interrupt_action}")
        if restart and not stop_run:
            raise SimulationError("Restarting a task, but it isn't stopping")
        return stop_run, restart

    @process
    def run(self, *, actor: "Actor") -> Generator[SimpyEvent | Process, Any, None]:
        """Execute the task.

        Args:
            actor (Actor): The actor using the task

        Returns:
            Generator[SimpyEvent, Any, None]
        """
        generator = self.task(actor=actor)
        self._proc = generator
        return_item = None
        stop_run = False
        restart = False
        come_back_to = False
        event_to_yield: Process | SimpyEvent
        while not stop_run:
            try:
                while True:
                    try:
                        if not come_back_to:
                            if return_item is None:
                                next_event = next(generator)
                            else:
                                next_event = generator.send(return_item)
                            # Allows processes to be yielded on inside an event
                            # This is dangerous
                            if isinstance(next_event, Process):
                                warn(
                                    f"Yielding a simpy.Process from {self}. "
                                    f"This is dangerous, take care. ",
                                    UserWarning,
                                )
                                event_to_yield = next_event
                            elif isinstance(next_event, BaseEvent):
                                event_to_yield = next_event.as_event()
                            else:
                                raise SimulationError(
                                    f"Unexpected yielded event type: {next_event}"
                                )
                        else:
                            come_back_to = False
                        return_item = yield event_to_yield
                        # TODO: test if the return_item is for a multi-event
                        # that way we can return it as a more useful object
                    except AttributeError as exc:
                        if "as_event" in exc.args[0]:
                            raise SimulationError("Task is yielding objects without `as_event`")
                        else:
                            raise exc
                    except StopIteration:
                        stop_run = True
                        break

            except Interrupt as interrupt:
                stop_run, restart = self._handle_interruption(
                    actor,
                    interrupt,
                    next_event,
                )
                if not stop_run:
                    come_back_to = True
                if restart:
                    generator = self.task(actor=actor)
                    self._proc = generator
                    return_item = None
                    stop_run = False
                    restart = False


class DecisionTask(Task):
    """A task used for decision processes."""

    DO_NOT_HOLD = False

    def task(self, *, actor: Any) -> TASK_TYPE:
        """Define the process this task follows."""
        raise SimulationError("No need to call `task` on a DecisionTask")

    def rehearse_decision(self, *, actor: Any) -> None:
        """Define the process this task follows."""
        raise NotImplementedError(NOT_IMPLEMENTED_MSG)

    def make_decision(self, *, actor: Any) -> None:
        """Define the process this task follows."""
        raise NotImplementedError(NOT_IMPLEMENTED_MSG)

    def rehearse(
        self,
        *,
        actor: REH_ACTOR,
        knowledge: dict[str, Any] | None = None,
        cloned_actor: bool = False,
        **kwargs: Any,
    ) -> REH_ACTOR:
        """Rehearse the task to evaluate its feasibility.

        Args:
            actor (Actor): The actor to rehearse with
            knowledge (Optional[dict[str, Any]], optional): Knowledge to add. Defaults to None.
            cloned_actor (bool, optional): If the actor is a clone or not. Defaults to False.
            kwargs (Any): Kwargs for rehearsal. Kept for consistency to the base class.

        Returns:
            Actor: Cloned actor after rehearsing this task.
        """
        knowledge = {} if knowledge is None else knowledge
        _old_env = self.env
        understudy = actor
        if not cloned_actor:
            understudy = self._clone_actor(actor, knowledge)
        self.env = understudy.env

        self._rehearsing = True

        self.rehearse_decision(actor=understudy)
        self.env = _old_env
        self._rehearsing = False
        return understudy

    @process
    def run(self, *, actor: "Actor") -> Generator[SimpyEvent, None, None]:
        """Run the decision task.

        Args:
            actor (Actor): The actor making decisions

        Yields:
            Generator[SimpyEvent, None, None]: Generator for SimPy event queue.
        """
        self.make_decision(actor=actor)
        assert isinstance(self.env, SimpyEnv)
        yield self.env.timeout(0.0)

    def run_skip(self, *, actor: "Actor") -> None:
        """Run the decision task with no clock reference.

        Task networks will use this method if SKIP_WAIT is True.

        Args:
            actor (Actor): The actor making decisions
        """
        self.make_decision(actor=actor)


class TerminalTask(Task):
    """A rehearsal-safe task that cannot exit, i.e., it is terminal.

    Note:
        The user can re-implement the `log_message` method to return a custom
        message that will be appended to the actor's log through its `log`
        method.
    """

    _time_to_complete: float = 1e24

    def log_message(self, *, actor: "Actor") -> str:
        """A message to save to a log when this task is reached.

        Args:
            actor (Actor): The actor using this task.

        Returns:
            str: A log message
        """
        return f"Entering terminal task: {self} on network {self._network_name}"

    def on_interrupt(self, *, actor: "Actor", cause: Any) -> InterruptStates:
        """Special case interrupt for terminal task.

        Args:
            actor (Actor): The actor
            cause (Any): Additional data sent to the interrupt.
        """
        raise SimulationError(
            f"Cannot interrupt a terminal task {self} on {actor}. Kwargs sent: {cause}"
        )
        return InterruptStates.END

    def task(self, *, actor: "Actor") -> TASK_TYPE:
        """The terminal task.

        It's just a long wait.

        Args:
            actor (Actor): The actor
        """
        log_message = self.log_message(actor=actor)
        actor.log(log_message)
        the_long_event = Event(rehearsal_time_to_complete=self._time_to_complete)
        res = yield the_long_event
        if res is not PLANNING_FACTOR_OBJECT:
            raise SimulationError(f"A terminal task completed on {actor}")



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""Geographical methods for intersections, distances, and locations."""

from .geo_types import (
    INTERSECTION_LOCATION_CALLABLE,
    LAT_LON,
    LAT_LON_ALT,
    EarthProtocol,
)
from .intersections import (
    CrossingCondition,
    get_intersection_locations,
)
from .spherical import Spherical
from .wgs84 import WGS84

__all__ = [
    "EarthProtocol",
    "Spherical",
    "WGS84",
    "get_intersection_locations",
    "LAT_LON",
    "LAT_LON_ALT",
    "CrossingCondition",
    "INTERSECTION_LOCATION_CALLABLE",
]
